"""
Tools for a metamodern web environment.

## User agent tools

Simple interface, simple automate.

## Web application framework

Simple interface, simple deploy.

"""

# TODO tidy up these imports

from dns import resolver as dns
from hstspreload import in_hsts_preload
import pendulum  # TODO XXX
from requests.exceptions import ConnectionError

from understory import mf
from understory.mkdn import render as mkdn
from understory import mm
from understory.mm import Template as template  # noqa
from understory.mm import templates  # noqa
from understory.uri import parse as uri

from . import agent
from .agent import *  # noqa
from . import braid
from .braid import *  # noqa
from . import framework
from .framework import *  # noqa
from .indie import *  # noqa
from .response import (Status,  # noqa
                       OK, Created, Accepted, NoContent, MultiStatus,
                       Found, SeeOther, PermanentRedirect,
                       BadRequest, Unauthorized, Forbidden, NotFound,
                       MethodNotAllowed, Conflict, Gone)
# from .tasks import run_queue

__all__ = ["dns", "in_hsts_preload", "mf", "mkdn", "mm", "template",
           "templates", "pendulum", "indieauth", "micropub", "microsub",
           "webmention", "websub", "run_queue", "uri", "Created",
           "ConnectionError"]
__all__ += agent.__all__ + braid.__all__ + framework.__all__  # noqa



"""Webmention receiver app and sending helper."""

from understory import web
from understory.web import tx


receiver = web.application("Webmention", mount_prefix="mentions",  # db=False,
                           mention_id=r"\w+")
templates = web.templates(__name__)


def send(source, targets, private=False) -> None:
    """Send a webmention."""
    # TODO ask archive.org to cache
    if source.startswith("/"):
        source = f"{tx.request.uri.scheme}://{tx.request.uri.netloc}/{source}"
    source = str(web.uri(source))
    target = str(web.uri(target))
    endpoint = web.discover_link(target, "webmention")
    if not endpoint:
        print(f"no webmention endpoint at {target}")
        return
    payload = {"source": source, "target": target}
    if private:
        code = web.nbrandom(32)
        # TODO tx.kv["mention-tokens"][":".join((path, str(target)))] = code
        payload["code"] = code
    store_mention(source, target)
    web.post(endpoint, data=payload)
    # TODO check if received and/or displayed
    # TODO !!! endpoint.post(payload)
    # TODO tx.db.update("mentions", what="data = ?", where="mention_id = ?",
    # TODO              vals=[mention, mention_id])


def receive(source, target) -> None:
    """Receive a webmention."""
    mention_id = store_mention(source, target)
    # mention_data = {}
    source_doc = tx.cache[source]
    source = source_doc.url.rstrip("/")
    mention = source_doc.mention(target).data  # TODO dict(..) inst .data
    tx.db.update("mentions", what="data = ?", where="mention_id = ?",
                 vals=[mention, mention_id])
    return

    # XXX source_data = mf.parse(source_doc.text, url=source)
    # XXX source_repr = mf.util.representative_hcard(source_data, source)
    # XXX if not source_repr:
    # XXX     raise web.BadRequest("No representative h-card found at source.")

    # XXX ZZZ path = web.uri(target).path
    # XXX ZZZ plural, _, resource_path = path.partition("/")
    # XXX ZZZ try:
    # XXX ZZZ     t = get_type(plural)
    # XXX ZZZ     resource = get_resource(t.s, resource_path)
    # XXX ZZZ except (TypeError, ResourceNotFound):
    # XXX ZZZ     resource = None

    # cache.put(source_url, parse=True)  # , html=source_doc.text)
    # source = interpret(source_url)
    # mention_data["source"] = source

    # # TODO ensure links are in page as with likes below
    # if "entry" in source:
    #     entry = source["entry"]
    #     entry_type = entry["-type"]
    #     if resource:
    #         if entry_type == "photo":
    #             print("photo post")
    #         elif entry_type == "associate":
    #             store("associations", path, source_url, mention_id)
    #         elif entry_type == "like":
    #             for like_of in entry["like-of"]:
    #                 if like_of["url"] == target:
    #                     store("likes", path, source_url, mention_id)
    #         elif entry_type == "vote":
    #             vote_on = entry["vote-on"][0]
    #             if vote_on["url"] == target:
    #                 store("votes", path, source_url, mention_id)
    #         elif entry_type == "note" and t.s == "code":
    #             t.utility(resource["-id"]).add_issue(source_url, entry,
    #                                                  mention_id)
    #         elif entry_type in ("note", "image") and \
    #                                t.s in ("note", "image"):
    #             if "in-reply-to" in entry:
    #                 for in_reply_to in entry["in-reply-to"]:
    #                     if in_reply_to["url"] == target:
    #                         store("replies", path, source_url, mention_id)
    #                 # bubble up mention if resource itself is a reply
    #                 if "in-reply-to" in resource:
    #                     send(path, resource["in-reply-to"])
    #     else:
    #         if entry_type == "follow":
    #             store("follows", path, source_url, mention_id)
    #         elif entry_type == "associate":
    #             store("associations", path, source_url, mention_id)
    # else:
    #     store("generic", path, source_url, mention_id)

    # mention_data["confirmed"] = time.time()
    # tx.kv["mentioned"][mention_id] = pickle.dumps(mention_data)

    # TODO tx.kv.db.publish(f"{tx.host.identity}:resources:{path}",
    # TODO                  f"<section class=h-entry>a {entry_type}</section>")
    #               str(tx.view.entry_template(entry, resource,
    #                                      t.view.type(entry, resource))))

    # # check if networking request
    # source_followees = source_data["rels"].get("followee", [])
    # source_followers = source_data["rels"].get("follower", [])
    # print()
    # print("ees", source_followees)
    # print("ers", source_followers)
    # if "code" in form:
    #     print("code", form.code)
    # print()
    # if "https://{}".format(tx.host.name) in source_followees:
    #     # TODO private webmention for initial follow
    #     # TODO public webmention for public follow
    #     # TODO check first private then public
    #     # TODO represent & reciprocate follow accordingly
    #     root_url = web.uri(source_repr["properties"]["url"][0], secure=True)
    #     person = get_person(root_url.minimized)
    #     rel = tx.db.select("person__person",
    #                        where="from_id = 1 AND to_id = ?",
    #                        vals=[person["id"]])[0]
    #     if rel:
    #         tx.db.update("person__person", what="private = 0",
    #                      where="from_id = 1 AND to_id = ?",
    #                      vals=[person["id"]])
    #         # TODO send return mention to notify if publicize
    #     return "okay"
    #     root_data = mf.parse(url=root_url)
    #     root_repr = mf.util.representative_hcard(root_data, root_url)
    #     if not root_repr:
    #         raise web.BadRequest("No representative h-card found "
    #                              "at source's root.")
    #     name = root_repr["properties"]["name"][0]
    #     try:
    #         email = root_repr["properties"]["email"][0]
    #     except IndexError:
    #         email = ""
    #     pubkey = ""
    #     person_feed = mf.util.interpret_feed(root_data, source_url)
    #     person_feed_license = root_data["rels"]["license"][0]
    #     add_person(name, root_url, email, pubkey, person_feed["name"],
    #                person_feed_license, "follower")


def store_mention(source, target):
    mention_id = web.nbrandom(9)
    tx.db.insert("mentions", mention_id=mention_id, source=source,
                 target=target)
    return mention_id


# def store(mention_type, path, source, mention_id) -> bool:
#     """
#     store a webmention
#
#     return True if it was added, false if it already exists
#
#     """
#     return tx.kv["mentions", path, mention_type].hset(source, mention_id)


def wrap(handler, app):
    """Ensure endpoint link in head of root document."""
    tx.db.define(mentions="""received DATETIME NOT NULL DEFAULT
                                 CURRENT_TIMESTAMP, mention_id TEXT,
                             data JSON, source TEXT, target TEXT""")
    # TODO sniff referer header for a mention
    # XXX print(f"REFERER: {tx.request.headers.get('Referer')}")
    yield
    if str(tx.response.headers.content_type) == "text/html":
        doc = web.parse(tx.response.body)
        try:
            head = doc.select("head")[0]
        except IndexError:
            pass
        else:
            head.append("<link rel=webmention href=/mentions>")
            tx.response.body = doc.html
        web.header("Link", '</mentions>; rel="webmention"', add=True)


receiver.wrap(wrap, "post")


def get_mentions(path):
    if path.startswith("/"):
        path = tx.origin + path
    return tx.db.select("mentions", where="target = ?", vals=[path])


@receiver.route(r"")
class Mentions:
    """."""

    def get(self):
        return templates.mentions(tx.db.select("mentions"))

    def post(self):
        form = web.form("source", "target")
        web.enqueue(receive, form.source, form.target)
        raise web.Accepted("webmention received")


@receiver.route(r"{mention_id}")
class Mention:
    """."""

    def get(self):
        """Details of the webmention, with status information in mf2."""
        mention = tx.db.select("mentions", where="mention_id = ?",
                               vals=[self.mention_id])
        return templates.mention(mention)

    def post(self):
        """Details of the webmention."""
        raise web.Accepted("webmention received")
        # XXX f"{tx.host.name}/mentions/{mention_id}")



"""

"""

# TODO scan for gun violence, etc.

# import os
import re

import lxml.html
# from lxml.html import builder as E

from understory.pkg import licensing
from understory import uri

# from . import hyphenator

__all__ = [  # "SmartQuotes",
           # # "SmallCapitals",
           # # "LicenseLink", "LicenseReference", "LicenseAutoReference",
           # # "OrgLink", "OrgReference", "OrgAutoReference",
           # # "PersonLink", "PersonReference", "PersonAutoReference",
           "ImageLink", "ImageReference", "ImageAutoReference",
           # "WikiLink",
           # "Reference", "AutoReference",
           # "AutoMagnet",
           # "Section", "Widow",
           "StrongEm", "Strong", "Em", "Small", "Code",

           "Link", "AutoLink", "WikiLink",
           "Mention", "Tag",

           # "Ampersand", "Copyright", "Ellipsis", "LigatureAE", "Heart",
           # "QuotationDash", "DblEmDash",
           # "InnerEmDash",
           "EmDash",
           # "InnerEnDash",
           "EnDash",
           # # "Hyphenate",
           ]


# STAGES = (__all__[:], __all__[:])
LINK_CONTAINER_RE = r"\[(?P<text>.+?)\]"


class Inline:

    """"""

    # def __init__(self, text):
    #     self.text = text

    # def str(self):
    #     return self.template.format(**self.parse(re.match(self.pattern,
    #                                                       self.text)))


class Link(Inline):

    """

        >>> link = Link("[Example](http://example.org/)")
        >>> str(link)
        u'<a href=""></a>'

    """

    split_pattern = r"\[.+?\]\(.+?\)"
    pattern = LINK_CONTAINER_RE + r"(?P<uri>\(.+?\))"

    def parse(self, match):
        given_url = match.group("uri")
        if given_url.startswith(("(#", "(/")):
            url = given_url[1:-1]
        else:
            if given_url.endswith(' "wikilink")'):
                url = uri.parse(self.parser.context)
                url.path = given_url[1:-12]
            else:
                url = uri.parse(given_url[1:-1])
            url = url.normalized
        text = match.group("text")
        return f'<a href="{url}">{text}</a>'
        # return E.A(text, href=url.normalized)


class AutoLink(Inline):

    """

    """

    split_pattern = r"https?://\S+"
    pattern = r"(?P<uri>https?://\S+)"

    def parse(self, match):
        url = uri.parse(match.group("uri"))
        self.parser.mentioned_urls.append(url)
        return f"<a href={url.normalized}>{url.minimized}</a>"
        # return E.A(url.minimized, href=url.normalized)


class WikiLink(Inline):

    """

    """

    pattern = r"\[\[(?P<page>.+?)\]\]"

    def parse(self, match):
        # TODO define wiki prefix in renderer config
        page = match.group("page")
        path = page.replace(" ", "")  # TODO CamelCase
        return f"<a href=/pages/{path}>{page}</a>"


class Mention(Inline):

    """

    """

    split_pattern = r"@[A-Za-z0-9.]+"
    pattern = r"@(?P<person>[A-Za-z0-9.]+)"

    def parse(self, match):
        if self.parser.child.tag == "a":
            self.parser.child.attrib["class"] = "h-card"
            return f"@{match.groups()[0]}"
        # TODO rel=tag for "entry" context
        person = match.group("person")
        self.parser.at_mentions.append(person)
        return f"<a class=h-card href=/people/{person}>{person}</a>"
        # return E.A(person, E.CLASS("h-card"), href=f"/people/{person}")


class Tag(Inline):

    """

    """

    split_pattern = r"#[A-Za-z0-9]+"
    pattern = r"#(?P<tag>[A-Za-z0-9]+)"

    def parse(self, match):
        # TODO rel=tag for "entry" context
        tag = match.group("tag")
        self.parser.tags.append(tag)
        return f"<a class=p-category href=/tags/{tag}>{tag}</a>"
        # return E.A(tag, E.CLASS("p-category"), href=f"/tags/{tag}")


# ----------------------------------------------------------------------


class SmartQuotes:

    """"""

    pattern = r".*"

    def parse(self, match):
        punct = r"""[!"#\$\%'()*+,-.\/:;<=>?\@\[\\\]\^_`{|}~]"""
        text = match.group(0)

        # Special case if the very first character is a quote followed by
        # punctuation at a non-word-break. Close the quotes by brute force:
        text = re.sub(r"""^'(?=%s\\B)""" % (punct,), r"""&#8217;""", text)
        text = re.sub(r"""^"(?=%s\\B)""" % (punct,), r"""&#8221;""", text)

        # Special case for double sets of quotes, e.g.:
        #   <p>He said, "'Quoted' words in a larger quote."</p>
        text = re.sub(r""""'(?=\w)""", """&#8220;&#8216;""", text)
        text = re.sub(r"""'"(?=\w)""", """&#8216;&#8220;""", text)

        # Special case for decade abbreviations (the '80s):
        text = re.sub(r"""\b'(?=\d{2}s)""", r"""&#8217;""", text)

        close_class = r"""[^\ \t\r\n\[\{\(\-]"""
        dec_dashes = r"""&#8211;|&#8212;"""

        # Get most opening single quotes:
        opening_single_quotes_regex = re.compile(r"""
                (
                    \s          |   # a whitespace char, or
                    &nbsp;      |   # a non-breaking space entity, or
                    --          |   # dashes, or
                    &[mn]dash;  |   # named dash entities
                    %s          |   # or decimal entities
                    &\#x201[34];    # or hex
                )
                '                 # the quote
                (?=\w)            # followed by a word character
                """ % (dec_dashes,), re.VERBOSE)
        text = opening_single_quotes_regex.sub(r"""\1&#8216;""", text)

        closing_single_quotes_regex = re.compile(r"""
                (%s)
                '
                (?!\s | s\b | \d)
                """ % (close_class,), re.VERBOSE)
        text = closing_single_quotes_regex.sub(r"""\1&#8217;""", text)

        closing_single_quotes_regex = re.compile(r"""
                (%s)
                '
                (\s | s\b)
                """ % (close_class,), re.VERBOSE)
        text = closing_single_quotes_regex.sub(r"""\1&#8217;\2""", text)

        # Any remaining single quotes should be opening ones:
        text = re.sub(r"""'""", r"""&#8216;""", text)

        # Get most opening double quotes:
        opening_double_quotes_regex = re.compile(r"""
                (
                    \s          |  # a whitespace char, or
                    &nbsp;      |  # a non-breaking space entity, or
                    --          |  # dashes, or
                    &[mn]dash;  |  # named dash entities
                    %s          |  # or decimal entities
                    &\#x201[34];   # or hex
                )
                "                  # the quote
                (?=\w)             # followed by a word character
                """ % (dec_dashes,), re.VERBOSE)
        text = opening_double_quotes_regex.sub(r"""\1&#8220;""", text)

        # Double closing quotes:
        closing_double_quotes_regex = re.compile(r"""
                #(%s)?  # character that indicates the quote should be closing
                "
                (?=\s)
                """ % (close_class,), re.VERBOSE)
        text = closing_double_quotes_regex.sub(r"""&#8221;""", text)

        closing_double_quotes_regex = re.compile(r"""
                (%s)   # character that indicates the quote should be closing
                "
                """ % (close_class,), re.VERBOSE)
        text = closing_double_quotes_regex.sub(r"""\1&#8221;""", text)

        # Any remaining quotes should be opening ones.
        text = re.sub(r'"', r"""&#8220;""", text)

        lsquo = r"""<span class="lsquo"><span>'</span></span>"""
        rsquo = r"""<span class="rsquo"><span>'</span></span>"""
        ldquo = r"""<span class="ldquo"><span>"</span></span>"""
        rdquo = r"""<span class="rdquo"><span>"</span></span>"""

        text = text.replace("&#8216;", lsquo)
        text = text.replace("&#8217;", rsquo)
        text = text.replace("&#8220;", ldquo)
        text = text.replace("&#8221;", rdquo)

        return text


class AutoMagnet(Inline):

    """

    """

    pattern = r"(?P<uri>magnet:.+?)"

    def parse(self, match):
        uri = match.group("uri")
        return '<a href="{0}">{0}</a>'.format(uri)


class Reference(Inline):

    """

        >>> reference = Reference("[Example][1]")
        >>> str(reference)
        u'<a href=""></a>'

    """

    pattern = LINK_CONTAINER_RE + r"\[(?P<reference_id>.+?)\]"
    template = '<a href="{uri}">{text}</a>'

    def parse(self, match):
        text = match.group("text")
        reference_id = match.group("reference_id")
        try:
            uri = self.parser.refs[reference_id]
        except KeyError:
            uri = ""  # XXX catch in parser and store as warning
            # raise ReferenceError("no reference for " + reference_id)
        return self.template.format(uri=uri, text=text)


class AutoReference:

    """

    """

    pattern = LINK_CONTAINER_RE
    template = '<a href="{uri}">{text}</a>'

    def parse(self, match):
        text = match.group("text")
        # XXX reference = lxml.html.fromstring(from_lga(text)).text_content()
        reference = lxml.html.fromstring(text).text_content()
        try:
            uri = self.parser.refs[reference]
        except KeyError:
            return match.group(0)
        return self.template.format(uri=uri, text=text)


class ImageLink(Link):

    """"""

    pattern = "!" + Link.pattern

    def parse(self, match):
        """"""
        text = match.group("text")
        uri = match.group("uri").strip("()")
        try:
            return '<img src="{0}" alt="{1}">'.format(uri, text)
        except KeyError:
            return match.group(0)


class ImageReference(Reference):

    """"""

    pattern = "!" + Reference.pattern

    def parse(self, match):
        text = match.group("text")
        reference_id = match.group("reference_id")
        try:
            uri = self.parser.refs[reference_id]
        except KeyError:
            return text
        return '<img src="{0}" alt="{1}">'.format(uri, text)


class ImageAutoReference(AutoReference):

    """"""

    pattern = "!" + AutoReference.pattern

    def parse(self, match):
        text = match.group("text")
        try:
            uri = self.parser.refs[text]
        except KeyError:
            return ""
        return '<img src="{0}" alt="{1}">'.format(uri, text)


# class LicenseLink(Link):
#
#     """"""


# class LicenseReference(Reference):
#
#     """"""


class LicenseAutoReference(AutoReference):

    """"""

    pattern = "%" + AutoReference.pattern

    def parse(self, match):
        alias = match.group(1)
        try:
            uri, name, version = licensing.get_license(alias)
        except licensing.LookupError:
            return alias
        anchor_template = '<a href="http://{}" rel=license>{} License</a>'
        return anchor_template.format(uri, name)


# class OrgLink(Link):
#
#     """"""
#
#     template = """"""


# class OrgReference(Reference):
#
#     """"""
#
#     template = '<span class="vcard h-card"><a class="fn p-fn '
#                'org h-org url u-url" href="{0}">{1}</a></span>'


# class OrgAutoReference(AutoReference):
#
#     """"""
#
#     pattern = "@@" + AutoReference.pattern
#     template = """<span class=h-card>{0}</span>"""
#
#     def parse(self, match):
#         text = match.group(1)
#         wrapper = '<span class="vcard h-card">{0}</span>'
#         try:
#             uri = self.parser.refs[text]
#         except KeyError:
#             return self.template.format('<span class="fn '
#                                         'org">{0}</span>'.format(text))
#         return wrapper.format('<a class="fn org url" '
#                               'href="{0}">{1}</a>'.format(uri, text))


# class PersonLink(Link):
#
#     """
#
#     """
#
#     pattern = "@" + Link.pattern
#     format_string = """<span class="vcard h-card">
#                        <a class="fn p-fn org h-org url u-url"
#                        href="{uri}">{text}</a>
#                        </span>"""


# class PersonReference(Reference):
#
#     """"""
#
#     template = """<span class="vcard h-card"><a class="fn """
#                """p-fn url u-url" href="{0}">{1}</a></span>"""


# class PersonAutoReference(AutoReference):
#
#     """"""
#
#     def parse(self, match):
#         """"""
#         text = match.group(1)
#         wrapper = u'<span class="vcard h-card">{0}</span>'
#         try:
#             uri = self.parser.refs[text]
#         except KeyError:
#             if "{" in text:
#                 text = text.replace("}", "</span>")
#                 z = re.compile(r"\{(?P<semantics>[\w,-]+):")
#
#                 def y(m):
#                     semantics = m.group("semantics").split(",")
#                     classes = " ".join(pr[2:] + " " + pr for pr in semantics)
#                     return u'<span class="{0}">'.format(classes)
#
#                 return wrapper.format(z.sub(y, text))
#             return wrapper.format(u'<span class="fn '
#                                   u'p-fn">{0}</span>'.format(text))
#         return wrapper.format(u'<a class="fn p-fn url u-url" '
#                               u'href="{0}">{1}</a>'.format(uri, text))


class Section:

    """"""

    pattern = r"(?P<symbol>SS?)\s+(?P<section>[\d.]+)"

    def parse(self, match):
        symbol = match.group("symbol")
        section = match.group("section")
        html = '<span class="x-section"><span>S</span></span>'
        return "&nbsp;".join((" ".join([html] * len(symbol)), section))


class Widow:

    """"""

    pattern = r"(?P<before>[\w\[\]]+)\s(?P<after>[\w'\[\]]+[\.,)]*\s*)$"

    def parse(self, match):
        return "&nbsp;".join((match.group("before"), match.group("after")))


# class SmallCapitals:
#
#     pattern = r"""
#         (?x)
#         \b
#         (?P<acronym>
#             AT&T | Q&A |
#             [A-Z\d][A-Z\d\./]{2,} |
#             [A-Z]{2,}
#         )
#         \b
#     """
#
#     def parse(self, match):
#         html = "<span class=x-small-caps>{0}</span>"
#         acronym = match.group("acronym")
#         # if self.roman_numeral_pattern.match(acronym).group("numeral"):
#         #     # html = "<span class=x-suggest-caps>{0}</span>"
#         #     return acronym
#         #     # if self.config["stage"] == "publish":
#         #     #     if acronym not in self.config["proofed"]:
#         #     #         return acronym
#         #     # else:
#         #     #     html = "<span class=x-suggest-caps>{0}</span>"
#         return html.format(acronym)
#
#     # Copyright (c) 2009, Mark Pilgrim, BSD License
#     roman_numeral_pattern = re.compile(r"""(?x)
#                                            (?P<numeral>
#                                                M{0,4}            # thousands
#                                                (CM|CD|D?C{0,3})  # hundreds
#                                                (XC|XL|L?X{0,3})  # tens
#                                                (IX|IV|V?I{0,3})  # ones
#                                            )""")


class StrongEm(Inline):

    # pattern = r"""\*{3}([\w\s\-.,'"]+)\*{3}"""
    pattern = r"""\*{3}(.+?)\*{3}"""

    def parse(self, match):
        text = match.group(1)
        return "<strong><em>{0}</em></strong>".format(text)


class Strong(Inline):

    # pattern = r"""\*{2}([\w\s\-.,'"]+)\*{2}"""
    pattern = r"""\*{2}(.+?)\*{2}"""

    def parse(self, match):
        text = match.group(1)
        return '<strong>{0}</strong>'.format(text)


class Em(Inline):

    # pattern = r"""\*([\w\s\-.,'"]+)\*"""
    pattern = r"""\*(.+?)\*"""

    def parse(self, match):
        text = match.group(1)
        return "<em>{0}</em>".format(text)


class Small(Inline):

    # pattern = r"""~([\w\s\-/<>%!.,'"]+)~"""
    pattern = r"""(?s)~(.+?)~"""

    def parse(self, match):
        text = match.group(1)
        return "<small>{0}</small>".format(text)


class Code(Inline):

    pattern = r"""(?s)`(.+?)`"""

    def parse(self, match):
        text = match.group(1)
        return "<code>{0}</code>".format(text)


class Ampersand(Inline):

    pattern = r" %!amp!% "

    def parse(self, match):
        return " <span class=x-amp>%!amp!%</span> "


class Copyright(Inline):

    pattern = r" \(c\) "

    def parse(self, match):
        return " &copy; "


class Ellipsis(Inline):

    pattern = r"\.\.\."

    def parse(self, match):
        # return "<span class=ellipsis><span>...</span></span>"
        return "&hellip;"


class QuotationDash(Inline):

    pattern = r"^---"

    def parse(self, match):
        return "<span class=quotation-dash><span>---</span></span>"


class DblEmDash(Inline):

    pattern = r"----"

    def parse(self, match):
        # return "<span class=em-dash><span>--</span></span>" * 2
        return "&emdash;" * 2


class EmDash(Inline):

    pattern = r"---"

    def parse(self, match):
        return "<span class=em-dash><span>&#x2014;</span></span>"


class InnerEmDash(Inline):

    pattern = r" --- "

    def parse(self, match):
        # return '<span class="em-dash inner"><span> --- </span></span>'
        return '&thinsp;&#x2014;&thinsp;'


class EnDash(Inline):

    pattern = r"--"

    def parse(self, match):
        return "<span class=en-dash><span>&#x2013;</span></span>"


class InnerEnDash(Inline):

    pattern = r" -- "

    def parse(self, match):
        # return '<span class="en-dash inner"><span>--</span></span>'
        return "&thinsp;&#x2013;&thinsp;"


class LigatureAE(Inline):

    pattern = r"(?i)(ae)"

    def parse(self, match):
        html = '<span class="x-ligature-ae x-{0}"><span>{1}</span></span>'
        cases = {"ae": "lower", "AE": "upper", "Ae": "upper"}
        m = match.group(1)
        return html.format(cases[m], m)


class Heart(Inline):

    pattern = r" <3 "

    def parse(self, match):
        return " &#x2764; "


# class Hyphenate(Inline):
#
#     # pattern = r"(?<!%!amp!%nbsp;)(?P<word>\w{4,}[\., \n])"
#     pattern = r"(?P<word>\w{4,}[\., \n])"
#
#     def parse(self, match):
#         hyphen_dict = os.path.dirname(__file__) + "/hyph_en_US.dic"
#         hyphenate = hyphenator.Hyphenator(hyphen_dict)
#         word = match.group("word")
#         sections = []
#         for section in word.split("-"):
#             sections.append(hyphenate.inserted(section).replace("-",
#                                                                 "&shy;"))
#         hyphenated = "-".join(sections)
#         start, _, final = hyphenated.rpartition("&shy;")
#         if len(final.strip()) < 3:  # or (final.endswith(".") and
#                                     # len(final.strip()) < 4):  # an-oth-er.
#             hyphenated = start + final
#         return hyphenated


gun_violence = {"Ammunition": "facts, evidence",
                "armed with the facts": "well-informed",
                "aim for": "hope to achieve",
                "ask point blank": "ask directly",
                "bite the bullet": "just do it",
                "bullet points": "specifics, key points",
                "bullet-proof": "invincible",
                "caught in the crossfire": "caught in the middle",
                "cheap shot": "unfair criticism",
                "dodged a bullet": "avoided trouble",
                "don't shoot the messenger": "not responsible",
                "even shot": "50/50 chance",
                "faster than a speeding bullet": "turbo speed",
                "'Fire Away!'": "'Get started!'",
                "firing blanks": "not succeeding",
                "fire back": "responding quickly",
                "going great guns": "succeeding beyond expectations",
                "gun it": "floor it",
                "gun shy": "reticent",
                "gunning for someone": "planning to harm",
                "half-cocked": "reckless",
                "high caliber": "exceptional",
                "hired gun": "paid specialist",
                "hold a gun to my head": "threaten",
                "hot shot": "show-off, braggart",
                "in the crosshairs": "scrutinized",
                "jumped the gun": "started too early",
                "'Just shoot me!'": "'I give up!'",
                "kill joy": "spoil sport",
                "killer instinct": "ruthless",
                "like shooting fish in a barrel": "too easy",
                "lock, stock and barrel": "all inclusive",
                "locked-and-loaded": "ready, prepared",
                "magic bullet": "perfect solution",
                "misfired": "erred",
                "missed the mark": "imperfect, fell short",
                "moving target": "hard to pin down",
                "outgunned": "outmatched",
                "on target": "accurate",
                "point blank": "direct, precise, simple",
                "point & shoot": "camera reference",
                "pot shots": "jabs",
                "pull the trigger": "get going",
                "quick on the trigger": "rash, hasty",
                "rapid fire": "quickly",
                "'Ready, aim, fire!'": "'Go!'",
                "riding shot gun": ("occupying the passenger seat",
                                    "protecting a companion"),
                "rifle through files": "search",
                "'She/he is a pistol!'": "'She/he has spunk!'",
                "shoot first, ask later": "impetuous",
                "shoot for": "try",
                "shoot for the moon": "set high goals",
                "shoot from the hip": "impulsive",
                "shoot me an email": "send me an email",
                "shoot off your mouth": "talk recklessly",
                "shoot out": "confrontation",
                "shoot the breeze": "chat, visit",
                "shot down my idea": "rejected",
                "shot in the dark": "wild guess",
                "shot myself in the foot": "made a mistake",
                "silver bullet": "perfect solution",
                "smoking gun": "proof beyond doubt",
                "'Son of a gun!'": "'Darn!'",
                "stick to my guns": "uncompromising",
                "straight shooter": "frank & honest",
                "sweating bullets": "really worried",
                "take aim": "focus",
                "take a shot at": "give it a try",
                "target market": "audience segment",
                "top gun": "recognized expert",
                "trigger a response": "elicit a response",
                "trigger alerts": "heads up, warnings",
                "trigger happy": "impulsive",
                "trip your trigger": "make you happy",
                "under fire": "being criticized",
                "under the gun": "feeling time pressured",
                "whole shooting match": "in total",
                "with guns blazing": "all-out effort",
                "worth a shot": "worth trying",
                "'You're killing me!'": "'You're too much!'"}



"""A simple Let's Encrypt interface (a la acme-tiny)."""

import acme_tiny
import sh


def generate_cert(domain, cert_dir, challenge_dir):
    """Generate a TLS certificate signed by Let's Encrypt for given domain."""
    account_key = cert_dir / "account.key"
    if not account_key.exists():
        sh.openssl("genrsa", "4096", _out=str(account_key))
    private_key = cert_dir / "domain.key"
    if not private_key.exists():
        sh.openssl("genrsa", "4096", _out=str(private_key))
    csr = cert_dir / "domain.csr"
    if not csr.exists():
        sh.openssl("req", "-new", "-sha256", "-key", private_key, "-subj",
                   f"/CN={domain}", _out=str(csr))
    with (cert_dir / "domain.crt").open("w") as fp:
        fp.write(acme_tiny.get_crt(account_key, csr, challenge_dir))
    csr.unlink()



"""IndieAuth client and server apps and sign-in helper."""

import base64
import hashlib

from understory import web
from understory.web import tx


root = web.application("IndieAuth", mount_prefix="auth",
                       db=False, client_id=r"[\w/.]+")
server = web.application("IndieAuthServer", mount_prefix="auth/sign-ins",
                         db=False, client_id=r"[\w/.]+")
client = web.application("IndieAuthClient", mount_prefix="auth/visitors",
                         db=False)
templates = web.templates(__name__)


def init_owner(name):
    """."""
    salt, scrypt_hash, passphrase = web.generate_passphrase()
    tx.db.insert("credentials", salt=salt, scrypt_hash=scrypt_hash)
    uid = str(web.uri(tx.origin))
    tx.pub.create("card", url=uid, uid=uid, name=name, nickname="me")
    tx.pub.create("feed", url="/all", uid="default", name="Posts")
    tx.pub.create("feed", url="/pages", uid="pages", name="Pages")
    tx.pub.create("entry", url="/", content="", channel=["pages"])
    tx.user.session = {"uid": uid, "name": name}
    tx.host.owner = tx.pub.get_card("me")
    return uid, passphrase


def wrap_server(handler, app):
    """Ensure server links are in head of root document."""
    tx.db.define(auths="""auth_id TEXT, initiated DATETIME NOT NULL DEFAULT
                              CURRENT_TIMESTAMP, revoked DATETIME,
                          code TEXT, client_id TEXT, client_name TEXT,
                          code_challenge TEXT, code_challenge_method TEXT,
                          redirect_uri TEXT, response JSON, token TEXT""",
                 credentials="""created DATETIME NOT NULL
                                    DEFAULT CURRENT_TIMESTAMP,
                                salt BLOB, scrypt_hash BLOB""")
    try:
        tx.user.is_owner = tx.user.session["uid"] == tx.host.owner["uid"][0]
    except (AttributeError, KeyError, IndexError):
        tx.user.is_owner = False
    if not tx.host.owner and tx.request.method == "GET":
        raise web.OK(templates.claim())
    passthrough = ("auth", "auth/sign-in", "auth/claim",
                   "auth/sign-ins/token", "auth/visitors/sign-in")
    if tx.request.uri.path.startswith(("auth", "pub", "sub")) and \
       tx.request.uri.path not in passthrough and \
       not tx.user.is_owner and \
       not tx.request.headers.get("Authorization"):  # TODO validate token
        raise web.Unauthorized(templates.unauthorized())
    yield
    if tx.request.uri.path == "" and tx.response.body:
        doc = web.parse(tx.response.body)
        base = "/auth/sign-ins"
        try:
            head = doc.select("head")[0]
        except IndexError:
            pass
        else:
            head.append(f"<link rel=authorization_endpoint href={base}>",
                        f"<link rel=token_endpoint href={base}/token>")
            tx.response.body = doc.html
        web.header("Link", f'<{base}>; rel="authorization_endpoint"', add=True)
        web.header("Link", f'<{base}/token>; rel="token_endpoint"', add=True)


def wrap_client(handler, app):
    """Ensure client database contains visitors table."""
    # TODO store User Agent and IP address
    tx.db.define(visitors="""account_created DATETIME NOT NULL DEFAULT
                                 CURRENT_TIMESTAMP, url TEXT, name TEXT,
                             email TEXT, access_token TEXT""")
    yield


def get_client(client_id):
    """Return the client name and author if provided."""
    # TODO FIXME unapply_dns was here..
    client = {"name": None, "url": web.uri(client_id).normalized}
    author = None
    if client["url"].startswith("https://addons.mozilla.org"):
        try:
            heading = tx.cache[client_id].dom.select("h1.AddonTitle")[0]
        except IndexError:
            pass
        else:
            client["name"] = heading.text.partition(" by ")[0]
            author_link = heading.select("a")[0]
            author_id = author_link.href.rstrip('/').rpartition('/')[2]
            author = {"name": author_link.text,
                      "url": f"https://addons.mozilla.org/user/{author_id}"}
    else:
        mfs = web.mf.parse(url=client["url"])
        for item in mfs["items"]:
            if "h-app" in item["type"]:
                properties = item["properties"]
                client = {"name": properties["name"][0],
                          "url": properties["url"][0]}
                break
            author = {"name": "NAME", "url": "URL"}  # TODO
    return client, author


@root.route(r"")
class AuthorizationRoot:
    """IndieAuth root manages either or both of server and client."""

    def get(self):
        return templates.root()


@root.route(r"sign-in")
class OwnerSignIn:
    """Sign in as the owner of the site."""

    def post(self):
        form = web.form("passphrase", return_to="/")
        credential = tx.db.select("credentials", order="created DESC")[0]
        if web.verify_passphrase(credential["salt"], credential["scrypt_hash"],
                                 form.passphrase.translate({32: None})):
            tx.user.session["uid"] = tx.host.owner["uid"][0]
            raise web.SeeOther(form.return_to)
        raise web.Unauthorized("bad passphrase")


@root.route(r"sign-out")
class OwnerSignOut:
    """Owner sign out."""

    def post(self):
        form = web.form(return_to="")
        print(form.return_to)
        tx.user.session = None
        raise web.SeeOther(f"/{form.return_to}")


@root.route(r"claim")
class ClaimOwner:
    """Claim site as owner."""

    def post(self):
        name = web.form("name").name
        # tx.user.session = None
        uid, passphrase = init_owner(name)
        return templates.claimed(" ".join(passphrase))


@server.route(r"")
class AuthorizationEndpoint:
    """IndieAuth server `authorization endpoint`."""

    def get(self):
        try:
            form = web.form("response_type", "client_id", "redirect_uri",
                            "state", "code_challenge", "code_challenge_method",
                            scope="")
        except web.BadRequest:
            clients = tx.db.select("auths", order="client_name ASC",
                                   what="DISTINCT client_id, client_name")
            active = tx.db.select("auths", where="revoked is null")
            revoked = tx.db.select("auths", where="revoked not null")
            return templates.authorizations(clients, active, revoked)
        client, developer = get_client(form.client_id)
        tx.user.session["client_id"] = form.client_id
        tx.user.session["client_name"] = client["name"]
        tx.user.session["redirect_uri"] = form.redirect_uri
        tx.user.session["state"] = form.state
        tx.user.session["code_challenge"] = form.code_challenge
        tx.user.session["code_challenge_method"] = form.code_challenge_method
        supported_scopes = ["create", "draft", "update", "delete",
                            "media", "profile", "email"]
        scopes = [s for s in form.scope.split() if s in supported_scopes]
        tx.user.is_authenticating = True
        return templates.signin(client, developer, scopes)

    def post(self):
        form = web.form("action", scopes=[])
        redirect_uri = web.uri(tx.user.session["redirect_uri"])
        if form.action == "cancel":
            raise web.Found(redirect_uri)
        code = f"secret-token:{web.nbrandom(32)}"
        s = tx.user.session
        decoded_code_challenge = base64.b64decode(s["code_challenge"]).decode()
        while True:
            try:
                tx.db.insert("auths", auth_id=web.nbrandom(3), code=code,
                             code_challenge=decoded_code_challenge,
                             code_challenge_method=s["code_challenge_method"],
                             client_id=s["client_id"],
                             client_name=s["client_name"],
                             redirect_uri=s["redirect_uri"],
                             response={"scope": " ".join(form.scopes)})
            except tx.db.IntegrityError:
                continue
            break
        redirect_uri["code"] = code
        redirect_uri["state"] = tx.user.session["state"]
        raise web.Found(redirect_uri)


@server.route(r"token")
class TokenEndpoint:
    """IndieAuth server `token endpoint`."""

    def post(self):
        try:
            form = web.form("action", "token")
            if form.action == "revoke":
                tx.db.update("auths", revoked=web.utcnow(), vals=[form.token],
                             where="""json_extract(response,
                                                   '$.access_token') = ?""")
                raise web.OK("")
        except web.BadRequest:
            pass
        form = web.form("grant_type", "code", "client_id",
                        "redirect_uri", "code_verifier")
        if form.grant_type != "authorization_code":
            raise web.Forbidden("only grant_type=authorization_code supported")
        auth = tx.db.select("auths", where="code = ?", vals=[form.code])[0]
        computed_code_challenge = \
            hashlib.sha256(form.code_verifier.encode("ascii")).hexdigest()
        if auth["code_challenge"] != computed_code_challenge:
            raise web.Forbidden("code mismatch")
        response = auth["response"]
        scope = response["scope"].split()
        if "profile" in scope:
            profile = {"name": tx.host.owner["name"][0]}
            if "email" in scope:
                try:
                    profile["email"] = tx.host.owner["email"][0]
                except KeyError:
                    pass
            response["profile"] = profile
        if scope and self.is_token_request(scope):
            response.update(access_token=web.nbrandom(12),
                            token_type="Bearer")
            response["me"] = (f"{tx.request.uri.scheme}://"
                              f"{tx.request.uri.netloc}")
        tx.db.update("auths", response=response,
                     where="code = ?", vals=[auth["code"]])
        web.header("Content-Type", "application/json")
        return response

    def is_token_request(self, scope):
        """Determine whether the list of scopes dictates a token request."""
        return bool(len([s for s in scope if s not in ("profile", "email")]))


@server.route(r"clients")
class Clients:
    """IndieAuth server authorized clients."""

    def get(self):
        clients = tx.db.select("auths", what="DISTINCT client_id, client_name",
                               order="client_name ASC")
        return templates.clients(clients)


@server.route(r"clients/{client_id}")
class Client:
    """IndieAuth server authorized client."""

    def get(self):
        auths = tx.db.select("auths", where="client_id = ?",
                             vals=[f"https://{self.client_id}"],
                             order="redirect_uri, initiated DESC")
        return templates.client(auths)


@client.route(r"")
class Visitors:
    """."""

    def get(self):
        return templates.visitors(dict(r) for r in tx.db.select("visitors"))


@client.route(r"sign-in")
class SignIn:
    """IndieAuth client sign in."""

    def get(self):
        try:
            form = web.form("me", return_to="/")
        except web.BadRequest:
            return templates.identify(tx.host.name)
        # XXX try:
        # XXX     rels = tx.cache[form.me].mf2json["rels"]
        # XXX except web.ConnectionError:
        # XXX     return f"can't reach {form.me}"
        # XXX auth_endpoint = web.uri(rels["authorization_endpoint"][0])
        # XXX token_endpoint = web.uri(rels["token_endpoint"][0])
        # XXX micropub_endpoint = web.uri(rels["micropub"][0])
        auth_endpoint = web.discover_link(form.me, "authorization_endpoint")
        token_endpoint = web.discover_link(form.me, "token_endpoint")
        micropub_endpoint = web.discover_link(form.me, "micropub")
        tx.user.session["auth_endpoint"] = str(auth_endpoint)
        tx.user.session["token_endpoint"] = str(token_endpoint)
        tx.user.session["micropub_endpoint"] = str(micropub_endpoint)
        client_id = web.uri(f"http://{tx.host.name}:{tx.host.port}")
        auth_endpoint["me"] = form.me
        auth_endpoint["client_id"] = tx.user.session["client_id"] = client_id
        auth_endpoint["redirect_uri"] = tx.user.session["redirect_uri"] = \
            client_id / "auth/visitors/authorize"
        auth_endpoint["response_type"] = "code"
        auth_endpoint["state"] = tx.user.session["state"] = web.nbrandom(16)
        code_verifier = tx.user.session["code_verifier"] = web.nbrandom(64)
        code_challenge = \
            hashlib.sha256(code_verifier.encode("ascii")).hexdigest()
        auth_endpoint["code_challenge"] = \
            base64.b64encode(code_challenge.encode("ascii"))
        auth_endpoint["code_challenge_method"] = "S256"
        auth_endpoint["scope"] = "create draft update delete profile email"
        tx.user.session["return_to"] = form.return_to
        raise web.SeeOther(auth_endpoint)


@client.route(r"authorize")
class Authorize:
    """IndieAuth client authorization."""

    def get(self):
        form = web.form("state", "code")
        if form.state != tx.user.session["state"]:
            raise web.BadRequest("bad state")
        payload = {"grant_type": "authorization_code",
                   "code": form.code,
                   "client_id": tx.user.session["client_id"],
                   "redirect_uri": tx.user.session["redirect_uri"],
                   "code_verifier": tx.user.session["code_verifier"]}
        response = web.post(tx.user.session["token_endpoint"],
                            headers={"Accept": "application/json"},
                            data=payload).json
        profile = response.get("profile", {})
        tx.db.insert("visitors", url=response["me"], name=profile.get("name"),
                     email=profile.get("email"),
                     access_token=response["access_token"])
        tx.user.session["uid"] = response["me"]
        raise web.SeeOther(tx.user.session["return_to"])


@client.route(r"sign-out")
class SignOut:
    """IndieAuth client sign out."""

    def post(self):
        token = tx.db.select("visitors", where="url = ?",
                             vals=[tx.user.session["uid"]])[0]["access_token"]
        web.post(tx.user.session["token_endpoint"],
                 data={"action": "revoke", "token": token})
        tx.user.session = None
        raise web.SeeOther("/")


def sign_in(user_url):
    """Initiate an IndieAuth sign-in (eg. micropub client)."""



"""
Render Python powered string templates.

>   But you're right, it is "yet another template language". And I'm not
>   going to apologize for it.

--- Aaron Swartz

Syntax
------

*   secure Python via [Templetor]
*   a strict superset dialect of [Markdown]
*   literate programming aided by [reStructuredText]
*   mathematical formulae a la [LaTeX]

Features
--------

*   spell check based upon GNU's Aspell (91 countries)
*   style conformance according to:

    *   [The Elements of Style]
    *   [The Elements of Typographic Style]

The following example demonstrates headings, block quotation, paragraphs,
ellipses, em dash, smart quotes, Microformats (hCard), lowercased figures,
title case, auto-reference and spell check.

(217 characters to 450+ including X codes.)

    >>> mana('''
    ... On nonsnse in prose  {nonsense}
    ... ===============================
    ...
    ... > ...
    ... > Beware the Jubjub bird, and shun
    ... > The frumious Bandersnatch!
    ... > ...
    ...
    ... --- @[Lewis Carroll], "Through the Looking-Glass,
    ...                        and What Alice Found There" 1872
    ... ''', stage='proof')  #doctest: +NORMALIZE_WHITESPACE
    '<h1 id=nonsense><a href=#nonsense>On <span class=x-spelling><span
    class=x-error>Nonsnse</span> (<span class=x-suggestions>Nonsense,
    Nonsenses, Nansen's, Nonsense's, Ninon's, Jonson's</span>)</span> in
    Prose</a></h1><blockquote><p>&#x; Be&shy;ware the Jub&shy;jub bird,
    and shun<br>The frumious Bandersnatch!&nbsp;&#x;</p></blockquote><p>&#x;
    <span class=h-card>Lew&shy;is Carroll</span>, &#x;Through the
    Look&shy;ing-Glass, and What Al&shy;ice Found There&#x;&nbsp;<span
    class=>1872</span></p>'

"""

from .templating import templates, build, Template, TemplatePackage

__all__ = ["templates", "build", "Template", "TemplatePackage"]



"""

"""

import inspect

from ..headers import Headers
from . import informational
from . import success
from . import redirection
from . import clienterror
from . import servererror
from .informational import *  # noqa
from .success import *  # noqa
from .redirection import *  # noqa
from .clienterror import *  # noqa
from .servererror import *  # noqa
from .util import Status

__all__ = ["Status"] + (informational.__all__ + success.__all__ +
                        redirection.__all__ + clienterror.__all__ +
                        servererror.__all__)


def parse(response):
    """
    return status, headers and body after parsing given response

    """
    lines = iter(response.splitlines())
    status = parse_status(next(lines))
    headers = Headers.from_lines(lines)
    return status, headers, "\n".join(lines)


def parse_status(line):
    """
    return a parsed

        >>> parse_status("HTTP/1.1 200 OK")
        ('1.1', OK('FIXME',))

    """
    raw_version, _, raw_status = line.partition(" ")
    version = raw_version.partition("/")[2]
    status = get_status(raw_status.partition(" ")[0])
    return version, status


def get_status(code):
    """
    return a `Status` object for given status `code`

    """
    if inspect.isclass(code) and issubclass(code, Status):
        return code("ASDF")  # FIXME
    code = int(code)
    for name, obj in globals().items():
        if not inspect.isclass(obj) or not issubclass(obj, Status):
            continue
        try:
            if code == int(inspect.getdoc(obj).partition(" ")[0]):
                return obj("ASDF")  # FIXME
        except ValueError:
            continue
    raise KeyError("cannot find a `Status` object for code {}".format(code))



"""Command line tools for the web."""

import json
from pprint import pprint
import textwrap

from understory import term
from understory import web

__all__ = ["main"]


main = term.application("web", web.__doc__)


@main.register()
class Serve:
    """Serve a web app."""

    def setup(self, add_arg):
        # add_arg("app", help="name of web application")
        pass

    def run(self, stdin, log):
        web.serve()

        # from pprint import pprint
        # pprint(web.get_apps())
        # for pkg, apps in web.get_apps().items():
        #     for name, _, ns, obj in apps:
        #         if self.app == name:
        #             web.serve(ns, obj)
        #             return 0
        # return 1


@main.register()
class MF:
    """Get microformats."""

    def setup(self, add_arg):
        add_arg("uri", help="address of the resource to GET and parse for MF")

    def run(self, stdin, log):
        pprint(web.get(self.uri).mf2json)
        return 0


@main.register()
class Micropub:
    """A Micropub editor."""

    # TODO media upload

    def setup(self, add_arg):
        add_arg("endpoint", help="address of the Micropub endpoint")
        add_arg("--type", default="entry", help="post type")
        add_arg("--token", default=None, help="IndieAuth bearer token")
        add_arg("--channel", nargs="*", help="add to given channel(s)")

    def run(self, stdin, log):
        properties = json.loads(stdin.read())
        try:
            properties["channel"].extend(self.channel)
        except KeyError:
            if self.channel:
                properties["channel"] = self.channel
        location, links = web.micropub.send(properties, endpoint=self.endpoint,
                                            h=self.type, token=self.token)
        print("Location:", location)
        print("Links:", links)
        return 0


@main.register()
class Microsub:
    """A Microsub reader."""

    def setup(self, add_arg):
        add_arg("endpoint", help="address of the Microsub endpoint")

    def run(self, stdin, log):
        return 0


@main.register()
class Braidpub:
    """A Braid publisher."""

    def setup(self, add_arg):
        add_arg("uri", help="address of the resource to publish")
        add_arg("range", help="content-range of the update in JSON Range")

    def run(self, stdin, log):
        patch_body = stdin.read()
        web.put(self.uri, headers={"Patches": "1"}, data=textwrap.dedent(f"""
            content-length: {len(patch_body)}
            content-range: {self.range}

            {patch_body}"""))
        return 0


@main.register()
class Braidsub:
    """A Braid subscriber."""

    def setup(self, add_arg):
        add_arg("uri", help="address of the resource to subscribe")

    def run(self, stdin, log):
        for version, parents, patches in web.subscribe(self.uri):
            if version:
                print("version:", version)
            if parents:
                print("parents:", parents)
            for patch_range, patch_body in patches:
                if patch_range:
                    print("range:", patch_range)
                print(patch_body)
                print()
            print()
        return 0



"""

"""

# TODO remove clean_indent in favor of dedent

from inspect import cleandoc as clean_indent
import re
from textwrap import dedent

import lxml.html
import lxml.html.builder as E
import pygments
import pygments.formatters
import pygments.lexers

from .titlecase import titlecase

__all__ = ["HTML", "Pre", "HashHeading", "ReSTHeading", "HorizontalRule",
           "OrderedList", "UnorderedList", "DefinitionList",
           "Blockquote", "Paragraph"]


class Block:

    """
    a generic block

    """

    def __init__(self, parser):
        self.parser = parser


class HTML(Block):

    """
    a block of raw HyperText

    """

    pattern = r"^<"

    def parse(self, block):
        """"""
        # if re.match(r"^</?[a-z0-9 =-]+>$", block):
        #     return E.PRE(block)  # TODO begin/end tags in blocks
        # XXX print(len(block))
        # XXX print(block)
        # XXX print()
        # XXX print()
        # XXX print()
        try:
            element = lxml.html.fromstring(block)
        except lxml.etree.ParserError:
            element = None
        return element
        # return lxml.html.fromstring(to_lga(block))


class Heading:

    """
    a generic heading

    """

    def process_heading(self, level, lines):
        """
        automatically titlecase so long as text doesn't end with " &"

        """
        sections = []
        for line in lines:
            if line.endswith("  "):
                line = line[:-2] + " <br>"
            sections.append(line)
        text = " ".join(sections)
        id_match = re.match(r".*\{(?P<id>\w[\w\d-]*[\w\d])\}$", text)
        if id_match:
            text = text.rpartition(" {")[0]
            id = id_match.group("id")
        else:
            id = text.lower().replace(" <br> ", "-").replace(" ", "-")
        text = text.rstrip()
        if text.endswith("&"):
            text = text.rstrip(" &")
            id = id.rstrip("-&")
        else:
            text = titlecase(text)
        html = "<h{0} id={1}>{2}</h{0}>"
        element = lxml.html.fromstring(html.format(level, id, text))
        # element= lxml.html.fromstring("<h{0}>{1}</h{0}>".format(level, text))
        # self.parser.toc.append((int(level), str(element.text_content()), id))
        return element


class HashHeading(Block, Heading):

    """
    a hash heading

    """

    pattern = r"^[#]{1,6}"

    def parse(self, block):
        """"""
        level = len(block.partition(" ")[0])
        lines = [line.lstrip("# ") for line in block.splitlines()]
        return self.process_heading(level, lines)


class ReSTHeading(Block, Heading):

    """
    a ReStructuredText (setext) heading

    """

    pattern = r"^(?s).*\n(=|-)+$"

    def parse(self, block):
        """"""
        level = "1" if block[-1] == "=" else "2"
        lines = block.splitlines()[:-1]
        return self.process_heading(level, lines)


class HorizontalRule(Block):

    """
    a horizontal rule

    """

    pattern = r"^\s*((\*|-|_|:)\s*){3,}$"

    def parse(self, block):
        """"""
        return E.HR()


class List:

    """
    a generic list

    """

    def parse(self, block):
        """"""
        list = self.list_type()
        for inner_block in re.split(self.list_pattern, block)[1:]:
            item = E.LI()
            self.parser.process_blocks(item, clean_indent(inner_block))
            list.append(item)
        return list


class OrderedList(Block, List):

    """
    an ordered list

    """

    pattern = r"^([\w#][.\)]\s{2}|\w{2}[.\)]\s)\S"
    list_pattern = r"(?m)^\w{1,2}\.\s+"
    list_type = E.OL


class UnorderedList(Block, List):

    """
    an unordered list

    """

    pattern = r"^(\*|\-|\+)[ ]{3}\S"
    list_pattern = r"(?m)^[\*\-\+][ ]{3}"
    list_type = E.UL


class DefinitionList(Block):

    """
    a definition list

    """

    pattern = r"^[\w\s]+\n:\s{3}\S"

    def parse(self, block):
        """"""
        list = E.DL()
        for inner_block in re.split(r"(?m)\n    \n(?=\w)", block):
            term_block, _, definition_block = inner_block.partition("\n:   ")
            for term in term_block.splitlines():
                list.append(E.DT(term))
            for def_block in definition_block.split("\n:   "):
                definition = E.DD()
                self.parser.process_blocks(definition,
                                           clean_indent(def_block))
                list.append(definition)
        return list


class Pre(Block):

    """
    a block of preformatted text

    """

    pattern = r"^[ ]{4}"

    def parse(self, block):
        """"""
        text = dedent(block)
        if text.startswith(">>> "):
            lexer = pygments.lexers.PythonConsoleLexer()
            lexer.add_filter("codetagify")
            formatter = pygments.formatters.HtmlFormatter(cssclass="doctest")
            code = pygments.highlight(text, lexer, formatter)
            html = lxml.html.fromstring(code)
        else:
            html = E.PRE(text)
        return html


class Blockquote(Block):

    """
    a block quotation

    The block quotation is used to denote a passage significant enough to
    stand alone.

        >   when & where?
        >
        >   >   Joe's @ Noon

        >>> blockquote = Blockquote('''
        ... >   when & where?
        ... >
        ... >   >   Joe's @ Noon''')
        >>> print(blockquote)
        <blockquote>
            <p>when &amp; where?</p>
            <blockquote>
                <p>Joe's @ Noon</p>
            </blockquote>
        </blockquote>

    """

    pattern = r"^>"

    def parse(self, block):
        """"""
        inner_blocks = "\n".join(l[2:] for l in block.splitlines())
        tree = E.BLOCKQUOTE()
        self.parser.process_blocks(tree, inner_blocks)
        return tree


class Paragraph(Block):

    """
    a paragraph

    The paragraph is the simplest block type as well as the fallback.

        Lorem ipsum dolor sit amet.

    Lorem ipsum dolor sit amet.

        >>> paragraph = Paragraph('''Lorem ipsum dolor sit amet.''')
        >>> print(paragraph)
        <p>Lorem ipsum dolor sit amet.</p>

    """

    pattern = r".*"

    def parse(self, block):
        """"""
        lines = []
        for line in block.splitlines():
            if line.endswith("  "):
                lines.append(line[:-2])
                lines.append(E.BR())
            else:
                lines.append(line + " ")
        return E.P(*lines)



"""
[Microsub][0] server and client apps.

[0]: https://indieweb.org/Microsub-spec

"""

from understory import web
from understory.web import tx

from .. import websub


server = web.application("MicrosubServer", mount_prefix="sub", db=False)
templates = web.templates(__name__)


def wrap_server(handler, app):
    """Ensure server links are in head of root document."""
    tx.db.define(following="""feed_id TEXT UNIQUE, url TEXT, callback_url TEXT,
                              added DATETIME NOT NULL
                                  DEFAULT CURRENT_TIMESTAMP""")
    tx.sub = LocalClient()
    yield
    if tx.request.uri.path == "" and tx.response.body:
        doc = web.parse(tx.response.body)
        try:
            head = doc.select("head")[0]
        except IndexError:
            pass
        else:
            head.append("<link rel=microsub href=/sub>")
            tx.response.body = doc.html
        web.header("Link", '</sub>; rel="microsub"', add=True)


class LocalClient:
    """."""

    def search(self, query):
        """
        Return a list of feeds associated with given query.

        If query is a URL, fetch it and look for feeds (inline and external).

        """
        # TODO Use canopy.garden for discovery.
        # TODO If query is a keyword, make suggestions.
        url = query
        if not url.startswith(("http://", "https://")):
            url = f"http://{url}"
        resource = tx.cache.add(url)
        feeds = []
        if resource.feed["entries"]:
            feed = {"type": "feed", "url": url, "name": "Unknown"}
            if resource.card:
                feed["name"] = resource.card["name"][0]
                try:
                    feed["photo"] = resource.card["photo"]
                except KeyError:
                    pass
            feeds.append(feed)
        for url in resource.mf2json["rels"].get("feed", []):
            feed = {"type": "feed", "url": url, "name": "Unknown"}
            subresource = tx.cache.add(url)
            if subresource.card:
                feed["name"] = subresource.card["name"][0]
                try:
                    feed["photo"] = subresource.card["photo"]
                except KeyError:
                    pass
            feeds.append(feed)
        return feeds

    def preview(self, url):
        """Return as much information about the URL as possible."""
        resource = tx.cache.add(url)
        items = []
        for entry in resource.feed["entries"]:
            item = {"type": "entry"}
            if "published" in entry:
                item["published"] = entry["published"]
            if "url" in entry:
                item["url"] = entry["url"]
            if "content" in entry:
                item["content"] = {"html": entry["content"],
                                   "text": entry["content-plain"]}
            if "category" in entry:
                item["category"] = entry["category"]
            if "photo" in entry:
                item["photo"] = entry["photo"]
            if "syndication" in entry:
                item["syndication"] = entry["syndication"]
            items.append(item)
        return {"items": items}

    def follow(self, url):
        """Start following the feed at given url."""
        feed_id = web.nbrandom(9)
        callback_url = f"{tx.origin}/hub/{feed_id}"
        tx.db.insert("following", url=url, feed_id=feed_id,
                     callback_url=callback_url)
        web.enqueue(websub.subscribe, url, callback_url)

    @property
    def following(self):
        """Return the feeds you're currently following."""
        return [{"type": "feed", "url": f["url"]}
                for f in tx.db.select("following")]


@server.route(r"")
class MicrosubServer:
    """Implement a Microsub endpoint."""

    def get(self):
        """Complete a Microsub transaction or return an activity summary."""
        try:
            form = web.form("action", channel="default")
        except web.BadRequest:
            following = tx.sub.following
            return templates.activity(following)
        if form.action == "follow":
            response = {"items": tx.sub.following}
        elif form.action == "timeline":
            response = {"items": []}
        web.header("Content-Type", "application/json")
        return response

    def post(self):
        """Complete a Microsub transaction."""
        form = web.form("action", channel="default")
        if form.action == "channels":
            pass
        elif form.action == "search":
            response = {"results": tx.sub.search(form.query)}
        elif form.action == "preview":
            response = tx.sub.preview(form.url)
        elif form.action == "follow":
            tx.sub.follow(form.url)
            response = {"items": tx.sub.following}
        elif form.action == "unfollow":
            pass
        elif form.action == "timeline":
            pass
        elif form.action == "mute":
            pass
        elif form.action == "unmute":
            pass
        elif form.action == "block":
            pass
        elif form.action == "unblock":
            pass
        web.header("Content-Type", "application/json")
        return response


@server.route(r"search")
class MicrosubServerSearch:
    """."""

    def get(self):
        """Return search results for given `q`."""
        form = web.form("q")
        return templates.search(form.q, tx.sub.search(form.q))


@server.route(r"preview")
class MicrosubServerPreview:
    """."""

    def get(self):
        """Return a preview of given `url`."""
        form = web.form("url")
        return templates.preview(form.url, tx.sub.preview(form.url))



"""

"""

from .util import Status

__all__ = ["Redirection", "MultipleChoices", "MovedPermanently",
           "Found", "SeeOther", "NotModified", "UseProxy",
           "TemporaryRedirect", "PermanentRedirect"]


class Redirection(Status):

    """3xx -- further action must be taken in order to complete the request"""


class MultipleChoices(Redirection):

    """
    300

    """


class MovedPermanently(Redirection):

    """
    301

    """


class Found(Redirection):

    """
    302

    """


class SeeOther(Redirection):

    """
    303

    """


class NotModified(Redirection):

    """
    304

    """


class UseProxy(Redirection):

    """
    305

    """


class SwitchProxy(Redirection):

    """
    306 *DEPRECATED*

    """


class TemporaryRedirect(Redirection):

    """
    307

    """


class PermanentRedirect(Redirection):

    """
    308

    """



# encoding: utf-8

"""

"""

import re


SMALL = "a(n|nd|s|t)?|but|by|en|for|if|in|of|on|or|the|to|vs?\.?|via"
PUNCT = r"""!"#$%&'‘()*+,\-./:;?@[\\\]_`{|}~"""
SMALL_WORDS = re.compile(r"^(%s)$" % SMALL, re.I)
INLINE_PERIOD = re.compile(r"[a-z][.][a-z]", re.I)
UC_ELSEWHERE = re.compile(r"[%s]*?[a-zA-Z]+[A-Z]+?" % PUNCT)
CAPFIRST = re.compile(r"^[%s]*?([A-Za-z])" % PUNCT)
SMALL_FIRST = re.compile(r"^([%s]*)(%s)\b" % (PUNCT, SMALL), re.I)
SMALL_LAST = re.compile(r"\b(%s)[%s]?$" % (SMALL, PUNCT), re.I)
SUBPHRASE = re.compile(r"([:.;?!][ ])(%s)" % SMALL)
APOS_SECOND = re.compile(r"^[dol]{1}['‘]{1}[a-z]+$", re.I)
ALL_CAPS = re.compile(r"^[A-Z\s%s]+$" % PUNCT)
UC_INITIALS = re.compile(r"^(?:[A-Z]{1}\.{1}|[A-Z]{1}\.{1}[A-Z]{1})+$")
MAC_MC = re.compile(r"^([Mm]a?c)(\w+)")


def titlecase(text):
    """

    """
    lines = re.split("[\r\n]+", text)
    processed = []
    for line in lines:
        all_caps = ALL_CAPS.match(line)
        words = re.split("[\t ]", line)
        tc_line = []
        for word in words:
            if all_caps:
                if UC_INITIALS.match(word):
                    tc_line.append(word)
                    continue
                else:
                    word = word.lower()
            if APOS_SECOND.match(word):
                word = word.replace(word[0], word[0].upper())
                word = word.replace(word[2], word[2].upper())
                tc_line.append(word)
                continue
            if INLINE_PERIOD.search(word) or \
               UC_ELSEWHERE.match(word):
                tc_line.append(word)
                continue
            if SMALL_WORDS.match(word):
                tc_line.append(word.lower())
                continue
            match = MAC_MC.match(word)
            if match:
                tc_line.append("%s%s" % (match.group(1).capitalize(),
                                         match.group(2).capitalize()))
                continue
            hyphenated = []
            for item in word.split("-"):
                _h = CAPFIRST.sub(lambda m: m.group(0).upper(), item)
                hyphenated.append(_h)
            tc_line.append("-".join(hyphenated))
        result = " ".join(tc_line)
        subfunc = lambda m: "{}{}".format(m.group(1), m.group(2).capitalize())
        result = SMALL_FIRST.sub(subfunc, result)
        result = SMALL_LAST.sub(lambda m: m.group(0).capitalize(), result)
        result = SUBPHRASE.sub(subfunc, result)
        processed.append(result)
    return "\n".join(processed)



"""

"""

import builtins
import glob
import imp
import importlib
import inspect
import os
import pathlib
import pkg_resources
import sys

from understory import mkdn
from understory import solarized

from . import parse

__all__ = ["templates", "build", "TemplatePackage", "Template",
           "CompiledTemplate", "TemplateResult"]


_builtin_names = ("dict", "enumerate", "float", "int", "bool", "list",
                  "long", "reversed", "set", "slice", "tuple", "xrange",
                  "abs", "all", "any", "callable", "chr", "cmp", "divmod",
                  "filter", "hex", "id", "isinstance", "iter", "len",
                  "sum", "max", "min", "oct", "ord", "pow", "range", "True",
                  "False", "None", "__import__",

                  "getattr", "object", "repr", "sorted", "basestring", "str",
                  "bytes", "type", "zip", "round", "dir", "next")


def templates(package, *template_paths, base_dir=None, **_globals):
    """

    """
    if not template_paths:
        path = pkg_resources.resource_filename(package, "templates")
        ns = importlib.import_module(package + ".templates")
        if base_dir:
            path = str(base_dir / path)
        return TemplatePackage(path, ns, **_globals)
    templates = []
    for template_path in template_paths:
        template_path = str(template_path)
        path = pkg_resources.resource_filename(package, template_path)
        if os.path.isdir(path):
            obj = templates(package, template_path)
        else:
            with open(path) as fp:
                template = fp.read()
            obj = Template(template, filename=template_path, globals=_globals)
        templates.append(obj)
    return templates


def build(directory):
    """
    compiles directory of templates to python code

    """
    for dirpath, dirnames, filenames in os.walk(directory):
        with open(os.path.join(dirpath, "__init__.py"), "w") as template:
            print("from mm.parse import *", file=template)
            print("from mm.templating import *", file=template)
            for dirname in dirnames[:]:
                if dirname.startswith("."):
                    dirnames.remove(dirname)
            if dirnames:
                print(file=template)
                for dirname in sorted(dirnames):
                    print("import {}".format(dirname), file=template)
            print(file=template)
            for fn in filenames:
                if fn.startswith((".", "__init__.py")) or fn.endswith("~"):
                    continue
                path = os.path.join(dirpath, fn)
                if "." in fn:
                    name = fn.split(".", 1)[0]
                else:
                    name = fn
                text = open(path).read()
                text = Template.normalize_text(text)
                code = Template.generate_code(text, path)
                code = code.replace("__template__", name, 1)
                print(file=template)
                print(code, end="\n\n", file=template)
                print('{0} = CompiledTemplate({0}, "{1}")'.format(name, path),
                      file=template)
                print("join_ = {}._join".format(name), file=template)
                print("escape_ = {}._escape".format(name), file=template)


def safestr(obj, encoding="utf-8"):
    r"""
    converts any given object to utf-8 encoded string

        >>> safestr("hello")
        "hello"
        >>> safestr("\u1234")
        "\xe1\x88\xb4"
        >>> safestr(2)
        "2"

    """
    if isinstance(obj, str):
        return obj
    # TODO elif isinstance(obj, unicode):
    # TODO     return obj.encode(encoding)
    # XXX elif hasattr(obj, "next") and hasattr(obj, "__iter__"):
    # XXX    return itertools.imap(safestr, obj)
    return str(obj)


html_entities = [("&", "amp"),  # must come first/last during quote/unquote
                 ("<", "lt"), (">", "gt"), ("'", "#39"), ('"', "quot")]


def htmlquote(text):
    r"""
    encode `text` for raw use in HTML

        >>> htmlquote("<'&\">")
        "&lt;&#39;&amp;&quot;&gt;"

    """
    text = str(text)
    for entity, code in html_entities:
        text = text.replace(entity, "&{};".format(code))
    return text


def websafe(val):
    r"""
    return a safe version of text for use in utf-8 encoded HTML, XHTML or XML

        >>> websafe("<'&\">")
        "&lt;&#39;&amp;&quot;&gt;"
        >>> websafe(None)
        ""
        >>> websafe("\u203d")
        "\u203d"
        >>> websafe("\xe2\x80\xbd")
        "\u203d"

    """
    if val is None:
        return ""
    # elif isinstance(val, str):
    #     val = val.decode("utf-8")
    # elif not isinstance(val, unicode):
    #     val = unicode(val)
    return htmlquote(val)
    # return htmlquote(str(val))
    # return htmlquote(bytes(val, "utf-8"))


def _import_file(path):
    """
    returns the distribution, module or package, imported from given path

    If `import_file` is called twice with the same module, the module
    is reloaded.

        >>> foo = _import_file("c:\\mylib.py")
        >>> bar = _import_file("relative_subdir/another.py")

    Based upon the original `import_file.py` copyright @[Yuval Greenfield][1]
    and released into the public domain.

    [1]: mailto:ubershmekel@gmail.com

    """
    if hasattr(os, "getcwdu"):
        # py2 returns question marks in os.path.realpath for ascii input (".")
        original_path = os.path.realpath(os.getcwdu())
    else:
        original_path = os.path.realpath(os.path.curdir)
    dst_path = os.path.dirname(path)
    if dst_path == "":
        dst_path = "."
    script_name = os.path.basename(path)
    if script_name.endswith(".py"):
        mod_name = script_name[:-3]
    else:
        mod_name = script_name
    os.chdir(dst_path)
    fhandle = None
    try:
        tup = imp.find_module(mod_name, ["."])
        module = imp.load_module(mod_name, *tup)
        fhandle = tup[0]
    finally:
        os.chdir(original_path)
        if fhandle is not None:
            fhandle.close()
    return module


class TemplatePackage:

    """
    a template renderer

        >>> templates = TemplatePackage()
        >>> templates.foo()  # doctest: +SKIP
        '<p>bar</p>'

    TemplatePackage are rendered recursively based upon directory layout.

    """

    def __init__(self, directory, ns=None, **kwglobals):
        """
        use the HTML found in in or around given `module`

        """
        self._directory = str(directory)
        if ns:
            self._ns = ns
        else:
            dir_path = pathlib.Path(directory)
            sys.path.insert(0, str(dir_path.parent))
            self._ns = importlib.import_module(dir_path.stem)
        self._globals = {k: v for k, v in self._ns.__dict__.items()
                         if k in getattr(self._ns, "__all__", [])}
        self._globals["mkdn"] = mkdn.render
        self._globals.update(kwglobals)
        self._cache = {}

    def _load_template(self, name):
        """
        load a template from root module directory

        An `html` subdirectory will be attempted if no template is found.

        """
        try:
            return self._cache[name]
        except KeyError:
            pass
        path = os.path.join(self._directory, name)
        if os.path.isdir(path):
            return TemplatePackage(path, self._ns, **self._globals)
        try:
            path = self._get_filename(path)
        except IndexError:
            raise AttributeError("No template named `{}`".format(name))
        with open(path) as f:
            template = f.read()
        compiled = Template(template, filename=path, globals=self._globals)
        if os.getenv("WEBCTX") != "dev":
            self._cache[name] = compiled
        return compiled

    __getattr__ = _load_template

    def _get_filenames(self, path):
        """"""
        return list(sorted(f for f in os.listdir(path) if not f.endswith("~")))

    def _get_filename(self, path):
        """"""
        return list(sorted(f for f in glob.glob(path + ".*")
                           if not (f == "__init__.py" or f.endswith("~"))))[0]

    def __dir__(self):
        return [os.path.split(p)[-1].split(".")[0] for p in
                self._get_filenames(os.path.join(self._directory))]


class Template:

    """

    """

    globals = {}
    builtins = dict((name, getattr(builtins, name)) for name in
                    _builtin_names if name in builtins.__dict__)

    def __init__(self, template, filename="<template>",
                 globals=None, builtins=None, extensions=None):
        self.extensions = extensions or []
        try:
            with template as fp:
                template = fp.read()
        except AttributeError:
            pass
        self._template = template
        text = Template.normalize_text(str(template))
        code = self.compile_template(text, filename)

        if globals is None:
            globals = self.globals
        if builtins is None:
            builtins = self.builtins

        self.filename = filename
        _globals = {"get_obj_name": lambda o: o.__name__,
                    "get_obj_docstring": inspect.getdoc,
                    "mkdn": mkdn.render}

        _globals.update(**globals)
        self._globals = _globals
        self._builtins = builtins
        if code:
            self.template = self._compile(code)  # XXX
        else:
            self.template = lambda: ""

    @staticmethod
    def normalize_text(text):
        """
        normalizes template text by correcting `\r\n`, tabs and BOM chars

        """
        text = text.replace("\r\n", "\n").replace("\r", "\n").expandtabs()
        if not text.endswith("\n"):
            text += "\n"
        BOM = "\xef\xbb\xbf"  # XXX support unicode? u"\ufeff"
        if isinstance(text, str) and text.startswith(BOM):
            text = text[len(BOM):]
        return text

    def __add__(self, other):
        try:
            new_text = other._template
        except AttributeError:
            new_text = other
        return Template(self._template + new_text)

    def _repr_html_(self):
        return solarized.highlight(self._template, ".html")

    def __call__(self, *args, **kwargs):
        """
        return rendered template

        As a side effect the current transaction's headers are updated
        according to content type of template.

        """
        types = {"txt": "text/plain",
                 "xml": "text/xml",
                 "html": "text/html",
                 "md": "text/x-markdown",
                 "css": "text/css",
                 "js": "text/javascript",
                 "json": "application/json",
                 "py": "text/x-python",
                 "svg": "image/svg+xml"}
        output = self.template(*args, **kwargs)
        ext = os.path.splitext(self.filename)[1][1:]
        output.content_type = types.get(ext, "text/html")
        # output.content_type = "text/html"
        return output

    @staticmethod
    def generate_code(text, filename, parser=None):
        parser = parser or parse.Parser()
        rootnode = parser.parse(text, filename)
        code = rootnode.emit(indent="").strip()
        return safestr(code)

    def create_parser(self):
        p = parse.Parser()
        for ext in self.extensions:
            p = ext(p)
        return p

    def compile_template(self, template_string, filename):
        code = Template.generate_code(template_string, filename,
                                      parser=self.create_parser())

        def get_source_line(filename, lineno):
            try:
                lines = open(filename).read().splitlines()
                return lines[lineno]
            except Exception:
                return None

        try:
            compiled_code = compile(code, filename, "exec")
        except SyntaxError as err:
            # display template line that caused the error along with traceback
            lineno = get_source_line(err.filename, err.lineno - 1)
            try:
                msg = "\n\nTemplate traceback:\n    File {}, line {}\n    {}"
                err.msg += msg.format(repr(err.filename), err.lineno - 5,
                                      lineno)
            except Exception:
                pass
            raise
        return compiled_code

    def _compile(self, code):
        env = self.make_env(self._globals or {}, self._builtins)
        exec(code, env)
        return env["__template__"]

    def make_env(self, globals, builtins):
        return dict(globals, __builtins__=builtins,
                    ForLoop=parse.ForLoop, TemplateResult=TemplateResult,
                    escape_=self._escape, join_=self._join)

    def _join(self, *items):
        return u"".join(items)

    def _escape(self, value, escape=False):
        if value is None:
            value = ""
        value = str(value)
        if escape:
            value = websafe(value)
        return value


class CompiledTemplate(Template):

    """

    """

    def __init__(self, f, filename):
        super(CompiledTemplate, self).__init__("", filename)
        self.template = f

    def compile_template(self, *a):
        return None

    def _compile(self, *a):
        return None


class TemplateResult(dict):

    """
    a dictionary like object for storing template output

    The result of a template execution is a `body` string and often a col-
    lection of attributes set using `$var: ...`. This class provides both a
    simple dictionary like interface for accessing these attributes as well
    as a string like interface for accessing the text output of the tem-
    plate.

    When the template is in execution, the output is generated part by part
    and those parts are combined at the end. Parts are added to the
    TemplateResult by calling the `extend` method and the parts are combined
    seemlessly when __body__ is accessed.

        >>> template = TemplateResult("hello, world", x="foo")
        >>> template
        <TemplateResult: 'hello, world' {'x': 'foo'}>
        >>> print(template)
        hello, world
        >>> template["x"]
        'foo'

        >>> template = TemplateResult()
        >>> template.extend([u"hello", u"world"])
        >>> template
        <TemplateResult: 'helloworld' {}>

    """

    __all__ = ["body"]

    def __init__(self, *a, **kw):
        self._body = None
        super().__init__(**kw)
        self._parts = []
        self.extend = self._parts.extend
        self.extend(a)

    @property
    def body(self):
        if self._body is None:
            self._body = "".join(str(p) if p else "" for p in self._parts)
        return self._body

    def __str__(self):
        return str(self.body).strip()

    # def __bytes__(self):
    #     return bytes(self.body, "utf-8")

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError("attr `{}` not found".format(name))

    def __add__(self, other):
        self._body = self.body + other.body
        return self

    def __bool__(self):
        return bool(self.body)

    # def __repr__(self):
    #     return "<TemplateResult: '{}' {}>".format(self, super().__repr__())



"""
A web application framework.

Strongly influenced by Aaron Swartz' "anti-framework" `web.py` this
library aims to cleanly abstract low-level web functionality through
a Pythonic API.

>   Think about the ideal way to write a web app. Write the
>   code to make it happen.

--- Aaron Swartz

"""

from base64 import b64encode, b64decode
import collections
# import errno
import getpass
import hashlib
import inspect
import os
import pkg_resources
import pathlib
import pprint
import re
import secrets
import shutil
import signal
import sys
import time
from time import sleep
import urllib

import Crypto.Random
import Crypto.Random.random
import gevent.pywsgi
import pendulum
import sh
import unidecode
import watchdog.events
import watchdog.observers

from understory import kv
from understory import mm
from understory.mm import Template
from understory import sql
from understory import term  # noqa NOTE to colorize print()

from ..agent import parse, apply_dns, cache
from ..response import (Status,  # noqa
                        OK, Created, Accepted, NoContent, MultiStatus,
                        Subscription,
                        Found, SeeOther, PermanentRedirect,
                        BadRequest, Unauthorized, Forbidden, NotFound,
                        MethodNotAllowed, Conflict, Gone)
from .letsencrypt import generate_cert
from .newmath import nbencode, nbdecode, nbrandom, nb60_re
from .passphrases import generate_passphrase, verify_passphrase
from .util import tx, header, json, JSONEncoder, shift_headings

import warnings  # mf2py's beautifulsoup's unspecified parser -- force lxml?
warnings.simplefilter("ignore")

random = secrets.SystemRandom()

__all__ = ["application", "serve", "anti_csrf", "form", "secure_form",
           "get_nonce", "get_token", "best_match", "sessions",
           "require_auth", "tx", "kv", "header",
           "Application", "Resource", "nbencode", "nbdecode", "nbrandom",
           "Template", "config_templates", "generate_cert",
           "get_integrity_factory", "utcnow", "JSONEncoder",
           "default_session_timeout", "textslug", "get_host_hash",
           "config_servers", "b64encode", "b64decode", "timeslug",
           "enqueue", "run_redis", "kill_redis", "get_apps", "nb60_re",
           "generate_passphrase", "verify_passphrase", "sleep", "random",
           "session_table_sql", "resume_session", "add_job_tables"]

kvdb = kv.db("web", ":", {"auth:secret": "string",
                          "auth:nonces": "set",
                          "reload-lock": "string",
                          "sessions:{session_id}": "string",
                          "sessions:{session_id}:data": "string"},
             session_id=r"[a-z0-9]{,65}")
config_templates = mm.templates(__name__)
methods = ["head", "get", "post", "put", "delete", "options", "patch",
           "propfind", "report"]
applications = {}
default_session_timeout = 86400


def ismethod(obj):
    """"""
    return (inspect.ismethod(obj) and obj.__name__.isupper() and
            obj.__name__.lower() in methods)


def application(name, *wrappers, prefix=r"", mount_prefix=r"",
                host=None, **path_args):
    if name in applications:
        app = applications[name]
        app.prefix = prefix  # FIXME does this have a side effect?
        app.add_wrappers(*wrappers)
        app.add_path_args(**path_args)
    else:
        app = Application(name, *wrappers, host=host, **path_args)
        app.reload_config()
        app.prefix = prefix
        app.mount_prefix = mount_prefix
        applications[name] = app
    return app


def get_host_hash(path):
    return hashlib.sha256(str(path.resolve()).encode("utf-8")).hexdigest()[:6]


def setup_servers(root, app, bot):
    """Initialize and configure servers (web, nginx, redis, supervisor)."""
    root = pathlib.Path(root).resolve()
    root_hash = get_host_hash(root)
    env = pathlib.Path(os.getenv("VIRTUAL_ENV"))
    root.mkdir()
    (root / "etc").mkdir()

    with (root / "etc/redis.conf").open("w") as fp:
        fp.write(str(config_templates.redis(root)))
    with (root / "etc/supervisor.conf").open("w") as fp:
        fp.write(str(config_templates.supervisor(root, root_hash,
                                                 getpass.getuser(),
                                                 env, app, bot)))
    with (root / "etc/nginx.conf").open("w") as fp:
        fp.write(str(config_templates.nginx(root_hash)))

    (root / "run").mkdir(parents=True)
    with (root / "etc/tor_stem_password").open("w") as fp:
        fp.write(str(random.getrandbits(100)))

    session_salt = Crypto.Random.random.randint(10**63, 10**64)
    cfg = {"root": str(root), "watch": str(env / "src"),
           "session": {"salt": str(session_salt)}}
    with (root / "etc/web.conf").open("w") as fp:
        json.dump(cfg, fp, indent=4, sort_keys=True)
    return root


# TODO rename
def config_servers(root, web_server_config_handler=None):
    """Update symlinks to config files and reload nginx."""
    root_hash = get_host_hash(root)
    # TODO FIXME XXX figure out what to do here...
    nginx_conf = pathlib.Path("/home/gaea/detritus/nginx/conf")
    sh.ln("-sf", (root / "etc/nginx.conf").resolve(),
          nginx_conf / "nginx.conf")
    if web_server_config_handler:
        web_server_config_handler(nginx_conf / "conf.d", root_hash)
    sh.sudo("ln", "-sf", (root / "etc/supervisor.conf").resolve(),
            "/etc/supervisor/conf.d/{}.conf".format(root_hash))


def serve():
    # root = pathlib.Path(app.cfg["root"])
    # watch = app.cfg.get("watch")
    # bg = False
    # if watch:
    #     bg = True

    def process(line):
        # TODO log everything, filter display to relevant
        print(line.rstrip())

    proc = sh.gunicorn(_bg=True, _err=process, _out=process)
    # if watch:
    event_handler = EventHandler(proc)
    observer = watchdog.observers.Observer()
    observer.schedule(event_handler, ".", recursive=True)
    observer.start()
    try:
        proc.wait()
    except KeyboardInterrupt:
        observer.stop()
    observer.join()
    proc.signal(signal.SIGQUIT)


def serve2(app_name, processes=2, threads=20, max_requests=100, lifespan=60):
    """Serve callable `app` at given `port`."""
    # TODO use file sockets specified as relative locations or in environ vars
    # master_fifo = os.path.expanduser("~/.web/{}-uwsgictl".format(app))

    # __import__(app)
    # for app_pkg, apps in get_apps().items():
    #     for app_name, app_instance in apps:
    #         if app == app_name:
    #             print(dir(app_pkg))
    app = applications[app_name]

    def process(line):
        # TODO log everything, filter display to relevant
        print(line.rstrip())

    root = pathlib.Path(app.cfg["root"])
    watch = app.cfg.get("watch")
    bg = False
    if watch:
        bg = True
    # XXX app_port = 8080
    proc = sh.uwsgi("-H", os.getenv("VIRTUAL_ENV"), "-w",
                    "{}.__web__:app".format(app_name),
                    "--uwsgi-socket", root / "run/web.sock",
                    "--chmod-socket=666",
                    # "--http-socket", ":{}".format(app_port),
                    "--logformat=%(method) %(host)%(uri) %(status)",
                    "--ignore-sigpipe",
                    "--max-requests", max_requests,
                    "--max-worker-lifetime", 60 * lifespan,
                    # "--master-fifo", master_fifo,
                    # "--enable-threads",
                    "-M", "-p", processes, "--async", threads, "--ugreen",
                    "--stats", root / "run/web-stats.sock",
                    # XXX "--stats", "127.0.0.1:1717", "--stats-http",
                    _bg=bg, _err=process)
    # logging.basicConfig(level=logging.INFO,
    #                     format='%(asctime)s - %(message)s',
    #                     datefmt='%Y-%m-%d %H:%M:%S')
    if watch:
        event_handler = EventHandler(proc)
        observer = Observer()
        observer.schedule(event_handler, str(watch), recursive=True)
        observer.start()
        try:
            proc.wait()
        except KeyboardInterrupt:
            observer.stop()
        observer.join()
        proc.signal(signal.SIGQUIT)


# def serve2(app_name):
#     """Serve callable `app` at given `port`."""
#     app = applications[app_name]
#
#     def process(line):
#         # TODO log everything, filter display to relevant
#         print(line.rstrip())
#
#     app_port = 8080
#     proc = sh.uwsgi("-H", os.getenv("VIRTUAL_ENV"), "-w",
#                     "{}.__web__:app".format(app_name),
#                     "--http-socket", ":{}".format(app_port),
#                     "--logformat=%(method) %(host)%(uri) %(status)",
#                     "--ignore-sigpipe",
#                     # "--master-fifo", master_fifo,
#                     # "--enable-threads",
#                     # XXX "--stats", "127.0.0.1:1717", "--stats-http",
#                     _bg=True, _err=process)
#     def quit():
#         proc.signal(signal.SIGQUIT)
#     # proc.wait()
#     return quit


class EventHandler(watchdog.events.FileSystemEventHandler):
    """Restart server when source directory changes."""

    def __init__(self, server_proc):
        self.server_proc = server_proc
        super(EventHandler, self).__init__()

    def on_modified(self, event):
        if event.src_path.endswith((".py", ".toml")):
            if kvdb:
                while kvdb["reload-lock"]:
                    time.sleep(2)
            self.server_proc.signal(signal.SIGHUP)

    # def on_moved(self, event):
    #     super(EventHandler, self).on_moved(event)
    #     # what = 'directory' if event.is_directory else 'file'
    #     # logging.info("Moved %s: from %s to %s", what, event.src_path,
    #     #              event.dest_path)

    # def on_created(self, event):
    #     super(EventHandler, self).on_created(event)
    #     # what = 'directory' if event.is_directory else 'file'
    #     # logging.info("Created %s: %s", what, event.src_path)

    # def on_deleted(self, event):
    #     super(EventHandler, self).on_deleted(event)
    #     # what = 'directory' if event.is_directory else 'file'
    #     # logging.info("Deleted %s: %s", what, event.src_path)

    # def on_modified(self, event):
    #     super(EventHandler, self).on_modified(event)
    #     # what = 'directory' if event.is_directory else 'file'
    #     # logging.info("Modified %s: %s", what, event.src_path)

    # on_moved = on_created = on_deleted = ...


def get_apps():
    """"""
    apps = collections.defaultdict(list)
    for ep in pkg_resources.iter_entry_points("web.apps"):
        try:
            handler = ep.load()
        except ModuleNotFoundError:
            print(f"couldn't load entry point: {ep}")
            continue
        try:
            raw_meta = ep.dist.get_metadata("PKG-INFO")
        except FileNotFoundError:
            raw_meta = ep.dist.get_metadata("METADATA").partition("\n\n")[0]
        ep.dist.metadata = dict(line.partition(": ")[0::2]
                                for line in raw_meta.splitlines())
        apps[ep.dist].append((ep.name, handler, ep.module_name, ep.attrs))
    return apps


def get_app(object_reference):
    """"""
    for ep in pkg_resources.iter_entry_points("web.apps"):
        if f"{ep.module_name}:{ep.attrs[0]}" == object_reference:
            return ep.name, ep.handler


def best_match(handlers, *args, **kwargs):
    """"""
    handler_types = [handler for handler, _ in handlers.items()]
    best_match = tx.request.headers.accept.best_match(handler_types)
    tx.response.headers.content_type = best_match
    return dict(handlers)[best_match](*args, **kwargs)


def get_integrity_factory(template_pkg):
    """Generate integrity hashes on the fly for local assets."""
    def handler(fn):
        static_assets_dir = pathlib.Path(template_pkg).parent.parent / "static"
        return sh.base64(sh.xxd(sh.sha256sum("-b", static_assets_dir / fn),
                                "-r", "-p")).strip()
    return handler


def utcnow():
    return pendulum.now("UTC")


_punct_re = r'[\t !"#$%&\'()*\-/<=>?@\[\\\]^_`{|},.]+'
_punct_split = re.compile(_punct_re).split


def textslug(title, delim="_", lower=True):
    """
    return a path slug containing a normalized version of `title`

    Defaults to delimiting words by "_" and enforcing all lowercase.
    Also makes intelligent path-friendly replacements (e.g. "&" to "and").

        >>> textslug("Surfing with @Alice & @Bob!")
        "surfing_with_alice_and_bob"

    """
    # TODO use formal normalization tooling:

    #   import unicodedata
    #   normalized = unicodedata.normalize("NFKC", title)

    #   title.encode("punycode")

    #   from confusable_homoglyphs import confusables
    #   bool(confusables.is_dangerous(title))

    result = []
    if lower:
        title = title.lower()
    title = title.replace(" & ", " and ")
    for word in _punct_split(title):
        result.extend(unidecode.unidecode(word).split())
    slug = delim.join(result)
    slug = slug.replace(":", "")  # XXX REMOVE WHEN MENTIONS REMOVED KV
    return slug


def timeslug(dt):
    """
    return a path slug containing date and time

    Date is encoded in standard YYYY/MM/DD form while the time is encoded
    as the NewBase60 equivalent of the day's centiseconds.

    """
    centiseconds = ((((dt.hour * 3600) + (dt.minute * 60) + dt.second) * 100) +
                    round(dt.microsecond / 10000))
    return "{}/{}/{}/{}".format(dt.year, dt.format("MM"), dt.format("DD"),
                                nbencode(centiseconds, 4))


def anti_csrf(handler):
    """"""
    # TODO csrf-ify forms with class name "x-secure"
    yield


def get_nonce():
    """"""
    return nbrandom(32)


def get_token(nonce):
    """
    return a token for given `secret` and `nonce`

    secret can be a string or a Redis instance in which key `secret`

    """
    secret = str(kvdb["auth:secret"])
    return hashlib.sha1(bytes(secret + nonce, "utf-8")).hexdigest()


def secure_form(*args, **kwargs):
    """"""
    args = list(args) + ["token", "nonce"]
    form = Form(*args, **kwargs)
    if not kvdb["auth:nonces"].add(form["nonce"]):
        raise OK("given `nonce` has already been used")
    if get_token(form["nonce"]) != form["token"]:
        raise OK("invalid `token` for given `nonce`")
    form.pop("token")
    form.pop("nonce")
    return form


class Form(dict):

    """

    """

    def __init__(self, *requireds, **defaults):
        _data = tx.request.body._data
        for required in requireds:
            if required not in _data:
                err_msg = "required `{}` not present in request"
                raise BadRequest(err_msg.format(required))
        super(Form, self).__init__(defaults)
        if isinstance(_data, str):
            return  # TODO FIXME what to do here? fix for Jupyter
        for key in _data.keys():
            if isinstance(_data[key], list):
                items = []
                for item in _data[key]:
                    if item.filename:
                        items.append(File(item.filename, item))
                    else:
                        items.append(item.value)
                self[key] = items
            else:
                if _data[key].filename:
                    value = File(key, _data[key])
                else:
                    value = _data.getfirst(key)
                if isinstance(defaults.get(key), list):
                    value = [value]
                self[key] = value

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError as e:
            raise AttributeError(e)

    def __setattr__(self, name, value):
        self[name] = value


form = Form


class File:

    """

    """

    def __init__(self, name, fileobj):
        self.name = name
        self.fileobj = fileobj

    def save(self, filepath, **options):
        """

        """
        filepath = pathlib.Path(filepath)
        # TODO handle required
        required = options.pop("required", False)
        if required:
            requireds = [required]
            if required == "jpg":
                requireds.append("jpeg")
            if not self.fileobj.filename.lower().endswith(tuple(requireds)):
                raise BadRequest("`{}` not `{}`".format(self.name, required))
        # self._ensure_dir_exists(dirpath)
        self.fileobj.file.seek(0)
        # while True:
        #     sugar = nbrandom(2)
        #     path = dirpath / self.name
        #     path = path.parent / "{}-{}{}".format(path.stem, sugar,
        #                                           path.suffix)
        if required:
            suffix = "." + required
        else:
            suffix = pathlib.Path(self.fileobj.filename).suffix
        filepath = filepath.with_suffix(suffix)
        if filepath.exists():
            raise FileExistsError("{} already exists".format(filepath))
        with filepath.open("wb") as fp:
            shutil.copyfileobj(self.fileobj.file, fp)
        return filepath

    # def _ensure_dir_exists(self, dirpath):
    #     try:
    #         os.makedirs(str(dirpath))
    #     except OSError as exc:
    #         if exc.errno == errno.EEXIST and dirpath.is_dir():
    #             pass
    #         else:
    #             raise


class Resource:

    """

    """

    def __init__(self, **kwargs):
        self.__dict__ = dict(kwargs)
        # XXX self._resources = ResourceData()
        # XXX for resource_name, value in kwargs.items():
        # XXX     table, _, column = resource_name.partition("_")
        # XXX     self._resources[table][column] = value

    def get_data(self):
        try:
            data = self.__dict__.pop("data")
        except KeyError:
            try:
                loader = self.load
            except AttributeError:
                def loader(): None
            data = loader()
        self.data = data

    def delegate(self, handler, *args, shift_headings=1, **kwargs):
        # TODO uhhmmm...
        kwargs.update(shift_headings=shift_headings)
        return handler()._get(self, *args, **kwargs)

    @classmethod
    def _get(resource, parent, **kwargs):
        """"""
        header_shift = kwargs.pop("shift_headings", None)
        if "_data" in parent.__dict__:
            parent._data.update(kwargs.get("data", {}))
        parent.__dict__.pop("data", None)  # NOTE magic name "data" is reserved
        kwargs.update(parent.__dict__)
        try:
            handler = resource(**kwargs)
            handler.get_data()
            content = handler.get()
        except NotFound:
            return "not found"
        if header_shift:
            try:
                content._body = shift_headings(str(content), header_shift)
            except AttributeError:
                pass
        return content

    def __contains__(self, item):
        return item in dir(self)


# XXX class ResourceData(collections.OrderedDict):
# XXX
# XXX     def __missing__(self, key):
# XXX         self[key] = collections.OrderedDict()
# XXX         return self[key]


def add_job_tables(sqldb):
    """"""
    sqldb.define(job_signatures="""module TEXT, object TEXT, args BLOB,
                                   kwargs BLOB, arghash TEXT,
                                   UNIQUE(module, object, arghash)""",
                 job_runs="""job_signature_id INTEGER,
                             job_id TEXT UNIQUE,
                             created DATETIME NOT NULL
                               DEFAULT(STRFTIME('%Y-%m-%d %H:%M:%f',
                                                'NOW')),
                             started DATETIME, finished DATETIME,
                             start_time REAL, run_time REAL,
                             status INTEGER, output TEXT""",
                 job_schedules="""job_signature_id INTEGER, minute TEXT,
                                  hour TEXT, day_of_month TEXT,
                                  month TEXT, day_of_week TEXT,
                                  UNIQUE(job_signature_id, minute,
                                         hour, day_of_month, month,
                                         day_of_week)""")


class Application:
    """
    A web application.

        >>> app = Application("example")
        >>> @app.wrap
        ... def contextualize(handler, app):
        ...     yield
        >>> @app.route(r"")
        ... class Greeting:
        ...     def get(self):
        ...         return "hello world"
        >>> response = app.get(r"")
        >>> response[0]
        '200 OK'
        >>> response[2][0]
        b'hello world'

    """

    def __init__(self, name, *wrappers, db=True, host=None, static=None,
                 icon=None, sessions=False, serve=False, **path_args):
        self.name = name
        self.wrappers = []
        self.pre_wrappers = []
        self.post_wrappers = []
        self.add_wrappers(*wrappers)
        self.host = host
        self.path_args = {}
        self.add_path_args(**path_args)
        self.mounts = []
        self.routes = []  # TODO use ordered dict

        if db:
            self.db = sql.db(f"web-{name}.db")
            self.cache = cache(db=self.db)

        class Database:
            def get(innerself):
                return tx.db.tables

        self.route(r"admin/db")(Database)
        self.add_path_args(table=r"\w+")

        class Table:
            def get(innerself):
                return pprint.pformat([dict(r)
                                       for r in tx.db.select(innerself.table)])

        self.route(r"admin/db/{table}")(Table)

        self.kv = kv.db(f"web-{name}", ":", {"jobs": "list"})
        if static:
            static_path = pkg_resources.resource_filename(static, "static")
            self.static_path = pathlib.Path(static_path)
        if sessions:  # TODO XXX isn't working in this wrap-order
            self.db.define(sessions=session_table_sql)
            self.wrap(resume_session, "pre")
        # for method in http.spec.request.methods[:5]:
        #     setattr(self, method, functools.partial(self.get_handler,
        #                                             method=method))
        # XXX try:
        # XXX     self.view = mm.templates(name, tx=tx)
        # XXX except ImportError:
        # XXX     try:
        # XXX         self.view = mm.templates(name + ".__web__", tx=tx)
        # XXX     except ImportError:
        # XXX         pass
        if icon:
            def insert_icon_rels(handler, app):
                yield
                if tx.response.status == "200 OK" and \
                   tx.response.headers.content_type == "text/html":
                    doc = parse(tx.response.body)
                    head = doc.select("head")[0]
                    head.append("<link rel=icon href=/icon.png>")
                    tx.response.body = doc.html
            self.wrap(insert_icon_rels, "post")

            if str(icon).endswith("="):  # TODO better test for b64
                payload = b64decode(icon)
            else:
                icon_path = pkg_resources.resource_filename(icon, "icon.png")
                payload = pathlib.Path(icon_path)

            class Icon:
                def get(innerself):
                    header("Content-Type", "image/png")
                    try:
                        with payload.open("rb") as fp:
                            return fp.read()
                    except AttributeError:
                        return payload
            self.route(r"icon.png")(Icon)
            self.route(r"favicon.ico")(Icon)
        if serve:
            self.serve(serve)

    def serve(self, port, host="127.0.0.1"):
        # must be called from custom jupyter kernel/head of your program:
        # XXX from gevent import monkey
        # XXX monkey.patch_all()
        class Log:
            def write(logline):
                (_, _, _, _, _, method, path,
                 _, status, _, duration) = logline.split()
                method = method.lstrip('"')
                duration = round(float(duration) * 1000)
                message = (f"<div class=httplog>"
                           f"<span>{self.host}</span>"
                           f"<span>{method}</span>"
                           f"<span>{path}</span>"
                           f"<span>{status}</span>"
                           f"<span>{duration}<span>ms</span></span>"
                           f"</div>")
                print(message)
        server = gevent.pywsgi.WSGIServer((host, port), self, log=Log)
        server.serve_forever()
        # TODO self.server = gevent.spawn(server.serve_forever)
        # TODO self.server.spawn()  # TODO start()?

    def reload_config(self, path=None):
        self.cfg = {}
        path = os.getenv("WEBCFG", path)
        try:
            with pathlib.Path(path).open() as fp:
                self.cfg = json.load(fp)
        except (FileNotFoundError, TypeError):
            pass

    def add_wrappers(self, *wrappers):
        self.wrappers.extend(wrappers)

    def add_path_args(self, **path_args):
        self.path_args.update({k: v.replace(r"\!", nb60_re)
                               for k, v in path_args.items()})

    def get(self, path):
        """"""
        env = {'HTTP_REFERER': 'http://localhost:22739/',
               'HTTP_HOST': 'lahacker.net:22739',
               'HTTP_CONNECTION': 'keep-alive',
               'HTTP_ACCEPT': 'image/png,image/*;q=0.8,*/*;q=0.5',
               'HTTP_USER_AGENT': 'Mozilla/5.0 (X11; Ubuntu; Linux i686; '
                                  'rv:21.0) Gecko/20100101 Firefox/21.0',
               'HTTP_COOKIE': 'session=99f81fd1f5e4c13b724043336f75'
                              'e81ee51d806979220cc8cebe51bc9d0de292',
               'HTTP_ACCEPT_LANGUAGE': 'en-US,en;q=0.5',
               'HTTP_ACCEPT_ENCODING': 'gzip, deflate',
               'PATH_INFO': path,
               'QUERY_STRING': '',
               'RAW_URI': '/',
               'REMOTE_ADDR': '127.0.0.1',
               'REMOTE_PORT': '44612',
               'REQUEST_METHOD': 'GET',
               'SCRIPT_NAME': '',
               'SERVER_PROTOCOL': 'HTTP/1.1',
               'SERVER_SOFTWARE': 'gunicorn/0.17.4',
               'SERVER_NAME': 'localhost',
               'SERVER_PORT': '22739',
               'wsgi.input': open("/dev/null"),
               'wsgi.url_scheme': 'http'}
        response = []

        def start_response(status, headers):
            response.append(status)
            response.append(headers)
        response.append(self(env, start_response))
        return tuple(response)

    def wrap(self, handler, when=None):
        """
        decorate a generator to run at various stages during the request

        """
        if when == "pre":
            self.pre_wrappers.append(handler)
        elif when == "post":
            self.post_wrappers.append(handler)
        else:
            self.wrappers.append(handler)
        return handler
        # def register(controller):
        #     path = path_template.format(**{k: "(?P<{}>{})".format(k, v) for
        #                                    k, v in self.path_args.items()})
        #     self.routes.append((path, controller))
        #     return controller
        # return register

    # @property
    # def resource(self):
    #     class Resource(type):
    #         def __new__(cls, name, bases, attrs):
    #             resource = type(name, bases, attrs)
    #           path = attrs["path"].format(**{k: "(?P<{}>{})".format(k, v) for
    #                                          k, v in self.path_args.items()})
    #             self.routes.append((path, resourcer))
    #             controller.__web__ = path
    #             return resource
    #     return Resource

    def route(self, path_template=None):
        """
        decorate a class to run when request path matches template

        """
        def register(controller):
            # TODO allow for reuse of parent_app's path_arg definitions
            # XXX try:
            # XXX     path_args = dict(self.parent_app.path_args,
            #                          **self.path_args)
            # XXX except AttributeError:
            # XXX     path_args = self.path_args
            templates = {k: "(?P<{}>{})".format(k, v) for
                         k, v in self.path_args.items()}
            try:
                path = path_template.format(**templates)
            except KeyError as err:
                raise KeyError("undefined URI fragment"
                               " type `{}`".format(err.args[0]))

            # TODO metaclass for `__repr__` -- see `app.routes`
            class Route(controller, Resource):

                __doc__ = controller.__doc__
                __web__ = path_template, templates
                handler = controller

            try:
                path = "/".join((self.prefix, path))
            except AttributeError:
                pass
            self.routes.append((path.strip("/"), Route))
            return Route
        return register

    def mount(self, *apps):
        """
        add an `application` to run when request path matches template

        """
        for app in apps:
            app.parent_app = self
            self.add_wrappers(*app.wrappers)  # TODO add pre and post wrappers
            path = app.mount_prefix.format(**{k: "(?P<{}>{})".format(k, v) for
                                              k, v in self.path_args.items()})
            self.mounts.append((path, app))

    def __repr__(self):
        return "<web.application: {}>".format(self.name)

    def __call__(self, environ, start_response):
        """The WSGI callable."""
        start = time.time()
        tx.request._contextualize(environ)
        tx.response._contextualize()
        path = tx.request.uri.path

        if path.startswith("static/"):
            asset_path = path.partition("/")[2]
            if asset_path.startswith((".", "/")):
                raise BadRequest("bad filename")
            asset = self.static_path / asset_path
            try:
                with asset.open() as fp:
                    content = fp.read()
            except FileNotFoundError:
                asset = pathlib.Path(__file__).parent / "static" / asset_path
                try:
                    with asset.open() as fp:
                        content = fp.read()
                except FileNotFoundError:
                    raise NotFound("file not found")
            content_types = {".css": "text/css",
                             ".js": "application/javascript"}
            # tx.response.headers.content_type = content_types[asset.suffix]
            header("Content-Type", content_types[asset.suffix])
            start_response("200 OK", tx.response.headers.wsgi)
            return [bytes(content, "utf-8")]

        try:
            tx.host._contextualize(environ, self, tx.request.headers.host.name,
                                   tx.request.headers.host.port)
        except AttributeError:
            raise NotFound("no hostname provided")
        tx.user._contextualize(environ)
        tx.log._contextualize()
        tx.response.headers.content_type = "text/plain"
        # XXX tx.app_name = self.name
        response_hooks = []

        def exhaust_hooks():
            for hook in response_hooks:
                try:
                    next(hook)
                except StopIteration:
                    pass

        try:
            tx.request.controller = self.get_controller(path)
            tx.response.headers.x_powered_by = "understory"

            for hook in self.pre_wrappers + self.wrappers + self.post_wrappers:
                if not inspect.isgeneratorfunction(hook):
                    msg = "`{}.{}` is not an iterator, give it a yield"
                    modname = getattr(hook, "__module__", "??")
                    raise TypeError(msg.format(modname, hook.__name__))
                _hook = hook(tx.request.controller, self)
                next(_hook)
                response_hooks.append(_hook)

            method = tx.request.method
            tx.response.status = "200 OK"
            if method == "GET":
                forced_method = Form().get("_http_method")
                if forced_method:
                    method = forced_method.upper()
            tx.request.controller.get_data()  # TODO FIXME err?
            body = self.get_handler(tx.request.controller, method)()
            if body is None:
                raise NoContent()
            if isinstance(body, tuple) and isinstance(body[0], dict):
                body = best_match(body[0], *body[1:])
            try:
                header("Content-Type", body.content_type)
            except AttributeError:
                pass
            tx.response.body = body
            # exhaust_hooks()  # TODO see below to pull hook exhaustion out
        except Status as exc:
            tx.response.status = str(exc)  # TODO exc.status
            tx.response.body = exc.body
            if exc.code == "201":
                tx.response.headers.location = exc.location
            if exc.code in ("301", "302", "303", "307", "308"):
                redirect_uri = apply_dns(str(tx.response.body))
                if redirect_uri.startswith("/"):  # relative HTTP(S) URL
                    tx.response.headers.location = \
                        urllib.parse.quote(redirect_uri)
                else:  # anything else (moz-extension, etc..)
                    tx.response.headers.location = redirect_uri
            if exc.code == "405":
                tx.response.headers.allow = ", ".join(dict(exc.allowed))
        except Exception:
            # raise  # NOTE leave to bubble naturally -- debug in terminal
            tx.response.status = "500 Internal Server Error"
            tx.response.headers.content_type = "text/html"
            if getattr(tx.user, "is_owner", False):
                body = config_templates.debug(*sys.exc_info())
            else:
                # TODO body = "Internal server error."
                body = config_templates.debug(*sys.exc_info())
                # TODO body = self.view.error.internal()
            tx.response.body = body
            # TODO logging
            # if os.environ["WWW_DEBUG"]:
            #     raise
        exhaust_hooks()  # here for mandatory wrappers (eg. session handling)
        # try:
        #     # exhaust_hooks()
        #     pass
        # except:
        #     print("BREAK IN THE HOOK")  # TODO provide debug
        #     raise
        try:
            header("Content-Type", tx.response.body.content_type)
        except AttributeError:
            pass

        duration = round(time.time() - start, 3)
        duration_color = None
        if duration < .1:
            pass
        elif duration < .5:
            duration_color = "y"
        else:
            duration_color = "r"
        duration = f"{duration:.3f}"
        if duration_color:
            duration = f"/{duration_color}/{duration}/X/"

        status = tx.response.status.partition(' ')[0]
        status_color = None
        if status.startswith("3"):
            status_color = "c"
        elif status.startswith("4"):
            status_color = "m"
        elif status.startswith("5"):
            status_color = "r"
        if status_color:
            status = f"/{status_color}/{status}/X/"

        method = tx.request.method
        host = tx.request.uri.host
        resource = tx.request.uri.path
        if tx.request.uri.query:
            resource += f"?{tx.request.uri.raw_query}"
        print(f"{duration} {status} {method: >6} {host}/{resource}")

        try:
            start_response(tx.response.status, tx.response.headers.wsgi)
        except OSError:  # websocket connection broken
            # TODO close websocket connection?
            return []
        if getattr(tx.response, "naked"):  # causes low-level gunicorn error
            return tx.response.body
        elif isinstance(tx.response.body, bytes):
            return [tx.response.body]
        elif tx.response.headers.get("content-type") == "application/json":
            # XXX return [bytes(json.dumps(tx.response.body), "utf-8")]
            return [bytes(JSONEncoder().encode(tx.response.body), "utf-8")]
        return [bytes(str(tx.response.body), "utf-8")]

    # def __gevent_call__(self, environ, start_response):
    #     """gevent's WSGI callable."""
    #     tx.request._contextualize(environ)
    #     tx.response._contextualize()
    #     path = tx.request.uri.path
    #     controller = self.get_controller(path)
    #     if tx.request.headers.get("Subscribe") == "keep-alive":
    #         handler = controller._subscribe
    #     else:
    #         handler = controller.get
    #     start_response("200 OK", [("Content-Type", "application/json")])
    #     return handler()

    def get_controller(self, path):
        """

        """
        # TODO softcode `static/` reference
        if path.endswith("/") and not path.startswith("static/"):
            raise PermanentRedirect("/" + path.rstrip("/"))

        for mount, app in self.mounts:
            m = re.match(mount, path)
            if m:
                controller = app.get_controller(path[m.span()[1]:].lstrip("/"))
                for k, v in m.groupdict().items():
                    setattr(controller, k, v)
                return controller

        class ResourceNotFound(Resource):
            def get(inner_self):
                raise NotFound("Resource not found")
                error = self.view.error
                # TODO recursively ascend app ancestors
                # try:
                #     error = self.parent_app.view.error
                # except AttributeError:
                #     error = self.parent_app.parent_app.view.error
                raise NotFound(error.resource_not_found())

        def get_resource():
            for pattern, resource in self.routes:
                if isinstance(resource, str):
                    mod = __import__(resource.__module__)
                    resource = getattr(mod, resource)
                match = re.match(r"^{}$".format(pattern),
                                 urllib.parse.unquote(path))
                if match:
                    return resource, match
            return ResourceNotFound, None

        resource, match = get_resource()  # TODO for-else or try-block instead?
        unquoted = {k: urllib.parse.unquote(v) for k, v
                    in match.groupdict().items() if v} if match else {}
        return resource(**unquoted)

    def get_handler(self, controller, method="get"):
        method = method.lower()
        try:
            return getattr(controller, method)
        except AttributeError:
            if method == "head":
                try:
                    return getattr(controller, "get")
                except AttributeError:
                    pass
            exc = MethodNotAllowed(f"{method.upper()} not allowed")
            exc.allowed = inspect.getmembers(controller, ismethod)
            raise exc


def sessions(**defaults):  # TODO XXX
    """
    returns an application hook for session handling using given redis `db`

    """
    def hook(handler, app):
        try:
            identifier = tx.request.headers["cookie"].morsels["session"]
        except KeyError:
            identifier = 0
        if ("sessions", identifier) not in kvdb:
            while True:
                secret = "{}{}{}{}".format(random.getrandbits(128),
                                           app.cfg["session"]["salt"],
                                           time.time(), tx.user.ip)
                secret_hash = hashlib.sha256(secret.encode("utf-8"))
                identifier = secret_hash.hexdigest()
                if kvdb["sessions", identifier].setnx("anonymous"):
                    break
        session_timeout = app.cfg["session"].get("timeout",
                                                 default_session_timeout)
        kvdb["sessions", identifier].expire(session_timeout)
        # FIXME user.session = kvdb.hgetall(kvdb(user.identifier, "data"))
        data = kvdb["sessions", identifier, "data"]
        if data:
            data = json.loads(str(data))
        else:
            data = Session(defaults)
        # XXX print(identifier, data)
        tx.user.session = data
        tx.user.identifier = identifier
        yield
        kvdb["sessions", identifier, "data"] = \
            JSONEncoder().encode(tx.user.session)
        kvdb["sessions", identifier, "data"].expire(session_timeout)
        # FIXME kvdb.delete("data")
        # FIXME if user.session:
        # FIXME     kvdb.hmset("data", **user.session)
        # XXX print(identifier, data, tx.user.identifier, tx.user.session)
        tx.response.headers["set-cookie"] = (("session", tx.user.identifier),)
        # ("Domain", tx.host.name))
    return hook


session_table_sql = """timestamp DATETIME NOT NULL
                           DEFAULT CURRENT_TIMESTAMP,
                       identifier TEXT NOT NULL UNIQUE,
                       data TEXT NOT NULL"""


def resume_session(handler, app):
    """."""
    # TODO monitor expiration (default_session_timeout)
    data = {}
    try:
        identifier = tx.request.headers["cookie"].morsels["session"]
    except KeyError:
        identifier = None
    else:
        try:
            session = tx.db.select("sessions", where="identifier = ?",
                                   vals=[identifier])[0]
        except IndexError:
            identifier = None
        else:
            try:
                data = json.loads(session["data"])
            except json.decoder.JSONDecodeError:
                identifier = None
    tx.user.session = data
    yield
    if tx.user.session:
        if identifier is None:
            salt = "abcdefg"  # FIXME
            secret = f"{random.getrandbits(64)}{salt}{time.time()}{tx.user.ip}"
            identifier = hashlib.sha256(secret.encode("utf-8")).hexdigest()
            tx.user.session.update(ip=tx.user.ip,
                                   ua=str(tx.request.headers["user-agent"]))
            # TODO FIXME add Secure for HTTPS sites (eg. canopy)!
            tx.response.headers["set-cookie"] = (("session", identifier),
                                                 ("path", "/"),
                                                 # "Secure",
                                                 "HttpOnly")
        tx.db.replace("sessions", identifier=identifier,
                      data=JSONEncoder().encode(tx.user.session))
    else:
        tx.response.headers["set-cookie"] = (("session", identifier),
                                             ("path", "/"), ("max-age", 0))


# TODO roles eg. owner, guest (personal context) && admin, user (service)
def require_auth(*roles):
    def decorate(func):
        def handler(*args, **kwargs):
            if tx.user.session["role"] not in roles:  # TODO role -> roles
                raise Unauthorized("no auth to access this resource")
            return func(*args, **kwargs)
        return handler
    return decorate


class Session(dict):

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            return super(Session, self).__getattr__(name)

    def __setattr__(self, name, value):
        self[name] = value


def run_redis(socket):
    sh.redis_server("--daemonize", "yes", "--unixsocket", socket, "--port", 0)


def kill_redis(socket):
    sh.redis_cli("-s", socket, "shutdown")


def enqueue(callable, *args, **kwargs):
    """
    append a function call to the end of the job queue

    """
    signature_id = get_job_signature(callable, *args, **kwargs)
    run_id = nbrandom(9)
    tx.db.insert("job_runs", job_id=run_id, job_signature_id=signature_id)
    # TODO add a "seen" column
    tx.kv["jobs"].append(run_id)
    return run_id


def get_job_signature(callable, *args, **kwargs):
    """
    return a job signature id creating a record if necessary

    """
    _module = callable.__module__
    _object = callable.__name__
    _args = json.dumps(args)
    _kwargs = json.dumps(kwargs)
    arghash = hashlib.sha256((_args + _kwargs).encode("utf-8")).hexdigest()
    try:
        job_signature_id = tx.db.insert("job_signatures", module=_module,
                                        object=_object, args=_args,
                                        kwargs=_kwargs, arghash=arghash)
    except tx.db.IntegrityError:
        job_signature_id = tx.db.select("job_signatures", what="rowid, *",
                                        where="""module = ? AND object = ? AND
                                                 arghash = ?""",
                                        vals=[_module, _object,
                                              arghash])[0]["rowid"]
    return job_signature_id



"""

"""


class Header:

    """"""

    def __init__(self, header):
        self.header = str(header).strip()
        self.parse()

    def parse(self):
        pass

    def __eq__(self, comparison):
        if isinstance(comparison, str):
            return self.header == comparison
        return False

    def __ne__(self, comparison):
        return not self == comparison

    def __contains__(self, what):
        return what in str(self.header)

    def __str__(self):
        return str(self.header)

    def __unicode__(self):
        return str(self.header)

    def __repr__(self):
        return str(self.header)
